#! /usr/bin/env python

"""
Module with utilities.
"""

from __future__ import division, print_function

__author__ = 'Carlos Alberto Gomez Gonzalez'
__all__ = []

import sys
import numpy as np

sep = '-' * 80


def check_array(input, dim=1, name=None):
    """ Checks the dimensionality of input. Returns it as a np.ndarray.
    """
    if name is None:
        name = 'Input'

    if dim == 1:
        msg = name + ' must be either a list or a 1d np.ndarray'
        if isinstance(input, (list, tuple)):
            input = np.array(input)
        if not isinstance(input, np.ndarray):
            raise TypeError(msg)
        if not input.ndim == 1:
            raise TypeError(msg)
    elif dim == 2:
        msg = name + ' must be an image or 2d np.ndarray'
        if not isinstance(input, np.ndarray):
            if not input.ndim == 2:
                raise TypeError(msg)
    elif dim == 3:
        msg = name + ' must be a cube or 3d np.ndarray'
        if not isinstance(input, np.ndarray):
            if not input.ndim == 3:
                raise TypeError(msg)
    elif dim == 4:
        msg = name + ' must be a cube or 4d np.ndarray'
        if not isinstance(input, np.ndarray):
            if not input.ndim == 4:
                raise TypeError(msg)
    return np.array(input)


def eval_func_tuple(f_args):
    """ Takes a tuple of a function and args, evaluates and returns result"""
    return f_args[0](*f_args[1:])                       


class redirect_output(object):
    """ Context manager for redirecting stdout/err to files"""
    def __init__(self, stdout='', stderr=''):
        self.stdout = stdout
        self.stderr = stderr

    def __enter__(self):
        self.sys_stdout = sys.stdout
        self.sys_stderr = sys.stderr

        if self.stdout:
            sys.stdout = open(self.stdout, 'w')
        if self.stderr:
            if self.stderr == self.stdout:
                sys.stderr = sys.stdout
            else:
                sys.stderr = open(self.stderr, 'w')

    def __exit__(self, exc_type, exc_value, traceback):
        sys.stdout = self.sys_stdout
        sys.stderr = self.sys_stderr




#! /usr/bin/env python

"""
Module with frame px resampling/rescaling functions.
"""
from __future__ import division, print_function

__author__ = 'Carlos Alberto Gomez Gonzalez, V. Christiaens @ ULg'
__all__ = ['frame_px_resampling',
           'cube_px_resampling',
           'cube_rescaling_wavelengths',
           'check_scal_vector']

import numpy as np
import warnings
try:
    import cv2
    no_opencv = False
except ImportError:
    warnings.warn("Opencv python bindings are missing.", ImportWarning)
    no_opencv = True

from scipy.ndimage.interpolation import geometric_transform, zoom
from ..var import frame_center


def cube_px_resampling(array, scale, imlib='ndimage', interpolation='bicubic',
                       verbose=True):
    """ Wrapper of ``frame_px_resampling`` for resampling the frames of a cube
    with a single scale factor. Useful when we need to upsample (upscaling) or
    downsample (pixel binning) a set of frames, e.g. an ADI cube.

    Parameters
    ----------
    array : array_like
        Input frame, 2d array.
    scale : int, float or tuple
        Scale factor for upsampling or downsampling the frames in the cube. If
        a tuple it corresponds to the scale along x and y.
    imlib : str optional
        See the documentation of the ``vip_hci.preproc.frame_px_resampling``
        function.
    interpolation : str, optional
        See the documentation of the ``vip_hci.preproc.frame_px_resampling``
        function.
    verbose : bool, optional
        Whether to print out additional info such as the new cube shape.

    Returns
    -------
    array_resc : array_like
        Output cube with resampled frames.

    """
    if array.ndim != 3:
        raise TypeError('Input array is not a cube or 3d array.')

    array_resc = []
    for i in range(array.shape[0]):
        imresc = frame_px_resampling(array[i], scale=scale, imlib=imlib,
                                     interpolation=interpolation)
        array_resc.append(imresc)

    array_resc = np.array(array_resc)

    if verbose:
        print("Cube successfully rescaled")
        print("New shape: {}".format(array_resc.shape))
    return array_resc


def frame_px_resampling(array, scale, imlib='ndimage', interpolation='bicubic',
                        verbose=False):
    """ Resamples the pixels of a frame wrt to the center, changing the size
    of the frame. If ``scale`` < 1 then the frame is downsampled and if
    ``scale`` > 1 then its pixels are upsampled.

    Parameters
    ----------
    array : array_like 
        Input frame, 2d array.
    scale : int, float or tuple
        Scale factor for upsampling or downsampling the frame. If a tuple it
        corresponds to the scale along x and y.
    imlib : {'ndimage', 'opencv'}, str optional
        Library used for image transformations. ndimage is the default.
    interpolation : str, optional
        For 'ndimage' library: 'nearneig', bilinear', 'bicuadratic', 'bicubic',
        'biquartic', 'biquintic'. The 'nearneig' interpolation is the fastest
        and the 'biquintic' the slowest. The 'nearneig' is the poorer
        option for interpolation of noisy astronomical images.
        For 'opencv' library: 'nearneig', 'bilinear', 'bicubic', 'lanczos4'.
        The 'nearneig' interpolation is the fastest and the 'lanczos4' the
        slowest and accurate.
    verbose : bool, optional
        Whether to print out additional info such as the new image shape.

    Returns
    -------
    array_resc : array_like 
        Output resampled frame.
    """
    if array.ndim != 2:
        raise TypeError('Input array is not a frame or 2d array')

    if isinstance(scale, tuple):
        scale_x, scale_y = scale
    elif isinstance(scale, (float, int)):
        scale_x = scale
        scale_y = scale
    else:
        raise TypeError('`scale` must be float, int or tuple')

    if imlib == 'ndimage':
        if interpolation == 'nearneig':
            order = 0
        elif interpolation == 'bilinear':
            order = 1
        elif interpolation == 'bicuadratic':
            order = 2
        elif interpolation == 'bicubic':
            order = 3
        elif interpolation == 'biquartic':
            order = 4
        elif interpolation == 'biquintic':
            order = 5
        else:
            raise TypeError('Scipy.ndimage interpolation method not recognized')

        array_resc = zoom(array, zoom=(scale_y, scale_x), order=order)

    elif imlib == 'opencv':
        if no_opencv:
            msg = 'Opencv python bindings cannot be imported. Install opencv or'
            msg += ' set imlib to ndimage'
            raise RuntimeError(msg)

        if interpolation == 'bilinear':
            intp = cv2.INTER_LINEAR
        elif interpolation == 'bicubic':
            intp= cv2.INTER_CUBIC
        elif interpolation == 'nearneig':
            intp = cv2.INTER_NEAREST
        elif interpolation == 'lanczos4':
            intp = cv2.INTER_LANCZOS4
        else:
            raise TypeError('Opencv interpolation method not recognized')

        array_resc = cv2.resize(array.astype(np.float32), (0, 0), fx=scale_x,
                                fy=scale_y, interpolation=intp)

    else:
        raise ValueError('Image transformation library not recognized')

    array_resc /= scale_y * scale_x

    if verbose:
        print("Image successfully rescaled")
        print("New shape: {}".format(array_resc.shape))

    return array_resc


def cube_rescaling_wavelengths(array, scaling_list, ref_xy=None, imlib='opencv',
                               interpolation='lanczos4', scaling_y=None,
                               scaling_x=None):
    """ Function to rescale a cube, frame by frame by a factor stored in
    ``scaling_list``, with respect to position (``ref_xy`` which by default
    is the center of the frames).
    
    Parameters
    ----------
    array : array_like 
        Input 3d array, cube.
    scaling_list : 1D-array
        Scale corresponding to each frame in the cube.
    ref_xy : float, optional
        Coordinates X,Y  of the point with respect to which the rescaling will be
        performed. By default the rescaling is done with respect to the center 
        of the frames; central pixel if the frames have odd size.
    imlib : {'opencv', 'ndimage'}, str optional
        Library used for image transformations. Opencv is faster than ndimage or
        skimage.
    interpolation : str, optional
        For 'ndimage' library: 'nearneig', bilinear', 'bicuadratic', 'bicubic',
        'biquartic', 'biquintic'. The 'nearneig' interpolation is the fastest
        and the 'biquintic' the slowest. The 'nearneig' is the poorer
        option for interpolation of noisy astronomical images.
        For 'opencv' library: 'nearneig', 'bilinear', 'bicubic', 'lanczos4'.
        The 'nearneig' interpolation is the fastest and the 'lanczos4' the
        slowest and accurate. 'lanczos4' is the default.
    scaling_y : 1D-array or list
        Scaling factor only for y axis. If provided, it takes priority on 
        scaling_list.
    scaling_x : 1D-array or list
        Scaling factor only for x axis. If provided, it takes priority on 
        scaling_list.
        
    Returns
    -------
    array_sc : array_like
        Resulting cube with rescaled frames.

    """
    def _scale_func(output_coords, ref_xy=0, scaling=1.0,
                    scale_y=None, scale_x=None):
        """
        For each coordinate point in a new scaled image (output_coords),
        coordinates in the image before the scaling are returned. This scaling
        function is used within geometric_transform which, for each point in the
        output image, will compute the (spline) interpolated value at the
        corresponding frame coordinates before the scaling.
        """
        ref_x, ref_y = ref_xy
        if scale_y is None:
            scale_y = scaling
        if scale_x is None:
            scale_x = scaling
        return (ref_y + (output_coords[0] - ref_y) / scale_y,
                ref_x + (output_coords[1] - ref_x) / scale_x)

    def _frame_rescaling(array, ref_xy=None, scale=1.0, imlib='opencv',
                         interpolation='lanczos4', scale_y=None, scale_x=None):
        """ Function to rescale a frame by a factor ``scale``, wrt a reference
        point which by default is the center of the frame (typically the exact
        location of the star). However, it keeps the same dimensions.

        Parameters
        ----------
        array : array_like
            Input frame, 2d array.
        ref_xy : float, optional
            Coordinates X,Y  of the point wrt which the rescaling will be
            applied. By default the rescaling is done with respect to the center
            of the frame.
        scale : float
            Scaling factor. If > 1, it will upsample the input array equally
            along y and x by this factor.
        scale_y : float
            Scaling factor only for y axis. If provided, it takes priority on
            scale parameter.
        scale_x : float
            Scaling factor only for x axis. If provided, it takes priority on
            scale parameter.

        Returns
        -------
        array_out : array_like
            Resulting frame.
        """
        if array.ndim != 2:
            raise TypeError('Input array is not a frame or 2d array.')

        if scale_y is None:
            scale_y = scale
        if scale_x is None:
            scale_x = scale

        outshape = array.shape
        if ref_xy is None:
            ref_xy = frame_center(array)

        if imlib == 'ndimage':
            if interpolation == 'nearneig':
                order = 0
            elif interpolation == 'bilinear':
                order = 1
            elif interpolation == 'bicuadratic':
                order = 2
            elif interpolation == 'bicubic':
                order = 3
            elif interpolation == 'biquartic':
                order = 4
            elif interpolation == 'biquintic':
                order = 5
            else:
                raise TypeError(
                    'Scipy.ndimage interpolation method not recognized')

            array_out = geometric_transform(array, _scale_func, order=order,
                                            output_shape=outshape,
                                            extra_keywords={'ref_xy': ref_xy,
                                                            'scaling': scale,
                                                            'scale_y': scale_y,
                                                            'scale_x': scale_x})

        elif imlib == 'opencv':
            if no_opencv:
                msg = 'Opencv python bindings cannot be imported. Install '
                msg += ' opencv or set imlib to skimage'
                raise RuntimeError(msg)

            if interpolation == 'bilinear':
                intp = cv2.INTER_LINEAR
            elif interpolation == 'bicubic':
                intp = cv2.INTER_CUBIC
            elif interpolation == 'nearneig':
                intp = cv2.INTER_NEAREST
            elif interpolation == 'lanczos4':
                intp = cv2.INTER_LANCZOS4
            else:
                raise TypeError('Opencv interpolation method not recognized')

            M = np.array([[scale_x, 0, (1. - scale_x) * ref_xy[0]],
                          [0, scale_y, (1. - scale_y) * ref_xy[1]]])
            array_out = cv2.warpAffine(array.astype(np.float32), M, outshape,
                                       flags=intp)

        else:
            raise ValueError('Image transformation library not recognized')

        array_out /= scale_y * scale_x
        return array_out

    ############################################################################
    if array.ndim != 3:
        raise TypeError('Input array is not a cube or 3d array')

    array_sc = []
    for i in range(array.shape[0]):
        array_sc.append(_frame_rescaling(array[i], ref_xy=ref_xy,
                                         scale=scaling_list[i], imlib=imlib,
                                         interpolation=interpolation,
                                         scale_y=scaling_y, scale_x=scaling_x))
    return np.array(array_sc)


def check_scal_vector(scal_vec):
    """ Function to check if the scaling list has the right format to avoid any
    bug in the pca algorithm, in the case of IFS data. All scaling factors
    should be >= 1 (i.e. the scaling should be done to match the longest
    wavelength of the cube).

    Parameter:
    ----------
    scal_vec: 1d array or list
        Vector with the scaling factors.

    Returns:
    --------
    scal_vec: array_like, 1d 
        Vector containing the scaling factors (after correction to comply with
        the condition >= 1).
    """
    if not isinstance(scal_vec, (list, np.ndarray)):
        raise TypeError('`Scal_vec` is neither a list or an np.ndarray')

    scal_vec = np.array(scal_vec)

    # checking if min factor is 1
    if scal_vec.min() != 1:
        scal_vec = 1/scal_vec
        scal_vec /= scal_vec.min()

    return scal_vec




#! /usr/bin/env python

"""
Module with various functions.
"""

from __future__ import division, print_function

__author__ = 'Carlos Alberto Gomez Gonzalez, O. Wertz'
__all__ = ['pp_subplots',
           'plot_surface',
           'get_fwhm',
           'lines_of_code']

import os
import numpy as np
from matplotlib.pyplot import (figure, subplot, show, colorbar, rc, axes,
                               Circle, savefig, close)
import matplotlib.colors as colors
from mpl_toolkits.mplot3d import Axes3D
from mpl_toolkits.axes_grid1 import make_axes_locatable
from .shapes import frame_center


def pp_subplots(*args, **kwargs):
    """ Wrapper for easy creation of pyplot subplots. It is convenient for 
    displaying VIP images in jupyter notebooks. 
    
    Parameters in **kwargs
    ----------------------
    angscale : bool
        If True, the axes are displayed in angular scale (arcsecs).
    angticksep : int
        Separation for the ticks when using axis in angular scale.
    arrow : bool
        To show an arrow pointing to input px coordinates.
    arrowalpha : float
        Alpha transparency for the arrow.
    arrowlength : int
        Length of the arrow, 20 px by default.
    arrowshiftx : int
        Shift in x of the arrow pointing position, 5 px by default.
    axis : bool
        Show the axis, on by default.
    circle : tuple or list of tuples
        To show a circle at given px coordinates, list of tuples.
    circlealpha : float or list of floats
        Alpha transparencey for each circle.
    circlelabel : bool
        Whether to show the coordinates of each circle.
    circlerad : int
        Radius of the circle, 6 px by default.
    cmap : str
        Colormap to be used, 'viridis' by default.
    colorb : bool
        To attach a colorbar, on by default.
    cross : tuple of float
        If provided, a crosshair is displayed at given px coordinates.
    crossalpha : float
        Alpha transparency of thr crosshair.
    dpi : int
        Dots per inch, for plot quality.
    getfig : bool
        Returns the matplotlib figure.
    grid : bool
        If True, a grid is displayed over the image, off by default.
    gridalpha : float
        Alpha transparency of the grid.
    gridcolor : str
        Color of the grid lines.
    gridspacing : int
        Separation of the grid lines in pixels.
    horsp : float
        Horizontal gap between subplots.
    label : str or list of str
        Text for annotating on subplots.
    labelpad : int
        Padding of the label from the left bottom corner.
    labelsize : int
        Size of the labels.
    log : bool
        Log colorscale.
    maxplots : int
        When the input (*args) is a 3d array, maxplots sets the number of
        cube slices to be displayed.
    pxscale : float
        Pixel scale in arcseconds/px. Default 0.01 for Keck/NIRC2.
    rows : int
        How many rows (subplots in a grid).
    save : str
        If a string is provided the plot is saved using this as the path.
    showcent : bool
        To show a big crosshair at the center of the frame.
    title : str
        Title of the plot(s), None by default.
    vmax : int
        For stretching the displayed pixels values.
    vmin : int
        For stretching the displayed pixels values.
    versp : float
        Vertical gap between subplots.

    """
    parlist = ['angscale', 'angticksep', 'arrow', 'arrowalpha', 'arrowlength',
               'arrowshiftx', 'axis', 'circle', 'circlealpha', 'circlerad',
               'circlelabel', 'cmap', 'colorb', 'cross', 'crossalpha', 'dpi',
               'getfig', 'grid', 'gridalpha', 'gridcolor', 'gridspacing',
               'horsp', 'label', 'labelpad', 'labelsize', 'log', 'maxplots',
               'pxscale', 'rows', 'save', 'showcent', 'title', 'vmax', 'vmin',
               'versp']
    
    for key in kwargs.keys():
        if key not in parlist:
            print("Parameter '{}' not recognized".format(key))
            print("Available parameters are: {}".format(parlist))

    # GEOM ---------------------------------------------------------------------
    num_plots = len(args)
    if num_plots == 1:
        if args[0].ndim == 3:
            data = args[0]
            if 'maxplots' in kwargs:
                maxplots = kwargs['maxplots']
            else:
                maxplots = 10
            num_plots = min(data.shape[0], maxplots)
        else:
            data = args
    elif num_plots > 1:
        data = args
        for i in range(num_plots):
            if args[i].ndim != 2:
                msg = "Wrong input. Must be either several 2d arrays (images) "
                msg += "or a single 3d array"
                raise TypeError(msg)

    if 'rows' in kwargs:
        if not isinstance(kwargs['rows'], int):
            raise TypeError
        else:
            rows = kwargs['rows']
    else:
        rows = 1

    if num_plots % rows == 0:
        cols = num_plots / rows
    else:
        cols = (num_plots / rows) + 1

    # CIRCLE -------------------------------------------------------------------
    if 'circle' in kwargs:
        if not isinstance(kwargs['circle'], list) and \
           isinstance(kwargs['circle'], tuple):
            show_circle = True
            coor_circle = [kwargs['circle']] * num_plots
        else:
            if not isinstance(kwargs['circle'][0], tuple):
                print("Circle must be a tuple (X,Y) or list of tuples (X,Y)")
                show_circle = False
            else:
                show_circle = True
                coor_circle = kwargs['circle']
    else:
        show_circle = False

    if 'circlerad' in kwargs:
        circle_rad = kwargs['circlerad']
    else:
        circle_rad = 6

    if 'circlealpha' in kwargs:
        circle_alpha = kwargs['circlealpha']
        if isinstance(kwargs['circlealpha'], (float, int)):
            circle_alpha = kwargs['circlealpha'] * len(coor_circle)
    else:
        if 'circle' in kwargs:
            if isinstance(kwargs['circle'], tuple):
                circle_alpha = 0.8
            elif isinstance(kwargs['circle'], list):
                circle_alpha = [0.8] * len(coor_circle)

    if 'circlelabel' in kwargs:
        circle_label = kwargs['circlelabel']
    else:
        circle_label = False

    # ARROW --------------------------------------------------------------------
    if 'arrow' in kwargs:
        if not isinstance(kwargs['arrow'], tuple):
            print("Arrow must be a tuple (X,Y)")
            show_arrow = False
        else:
            coor_arrow = kwargs['arrow']
            show_arrow = True
    else:
        show_arrow = False

    if 'arrowshiftx' in kwargs:
        arrow_shiftx = kwargs['arrowshiftx']
    else:
        arrow_shiftx = 5

    if 'arrowlength' in kwargs:
        arrow_length = kwargs['arrowlength']
    else:
        arrow_length = 20

    if 'arrowalpha' in kwargs:
        arrow_alpha = kwargs['arrowalpha']
    else:
        arrow_alpha = 0.8

    # LABEL --------------------------------------------------------------------
    if 'label' in kwargs:
        label = kwargs['label']
        if len(label) != num_plots:
            print("Label list does not have enough items")
            label = None
    else:
        label = None

    if 'labelsize' in kwargs:
        labelsize = kwargs['labelsize']
    else:
        labelsize = 12

    if 'labelpad' in kwargs:
        labelpad = kwargs['labelpad']
    else:
        labelpad = 5

    # GRID ---------------------------------------------------------------------
    if 'grid' in kwargs:
        grid = kwargs['grid']
    else:
        grid = False

    if 'gridcolor' in kwargs:
        grid_color = kwargs['gridcolor']
    else:
        grid_color = '#f7f7f7'

    if 'gridspacing' in kwargs:
        grid_spacing = kwargs['gridspacing']
    else:
        grid_spacing = None

    if 'gridalpha' in kwargs:
        grid_alpha = kwargs['gridalpha']
    else:
        grid_alpha = 0.4

    # VMAX-VMIN ----------------------------------------------------------------
    if 'vmax' in kwargs:
        if isinstance(kwargs['vmax'], (tuple, list)):
            if len(kwargs['vmax']) != num_plots:
                print("Vmax does not list enough items, setting all to None")
                vmax = [None]*num_plots
            else:
                vmax = kwargs['vmax']
        else:
            vmax = [kwargs['vmax']]*num_plots
    else:
        vmax = [None]*num_plots

    if 'vmin' in kwargs:
        if isinstance(kwargs['vmin'], (tuple, list)):
            if len(kwargs['vmin']) != num_plots:
                print("Vmin does not list enough items, setting all to None")
                vmin = [None]*num_plots
            else:
                vmin = kwargs['vmin']
        else:
            vmin = [kwargs['vmin']]*num_plots
    else:
        vmin = [None]*num_plots

    # CROSS --------------------------------------------------------------------
    if 'cross' in kwargs:
        if not isinstance(kwargs['cross'], tuple):
            print("Crosshair must be a tuple (X,Y)")
            show_cross = False
        else:
            coor_cross = kwargs['cross']
            show_cross = True
    else:
        show_cross = False
    if 'crossalpha' in kwargs:
        cross_alpha = kwargs['crossalpha']
    else:
        cross_alpha = 0.4

    # AXIS - ANGSCALE ----------------------------------------------------------
    if 'angticksep' in kwargs:
        angticksep = kwargs['angticksep']
    else:
        angticksep = 50
    if 'pxscale' in kwargs:
        pxscale = kwargs['pxscale']
    else:
        pxscale = 0.01  # default for Keck/NIRC2
    if 'angscale' in kwargs:
        angscale = kwargs['angscale']
        if angscale:
            print("`Pixel scale set to {}`".format(pxscale))
    else:
        angscale = False
    if 'axis' in kwargs:
        show_axis = kwargs['axis']
    else:
        show_axis = True
    # --------------------------------------------------------------------------

    show_center = False
    if 'showcent' in kwargs and kwargs['showcent']:
        show_center = True

    if 'getfig' in kwargs:
        getfig = kwargs['getfig']
    else:
        getfig = False

    if 'save' in kwargs and isinstance(kwargs['save'], str):
        save = True
        savepath = kwargs['save']
    else:
        save = False

    # Defaults previously used: 'magma','CMRmap','RdBu_r'
    if 'cmap' in kwargs:
        custom_cmap = kwargs['cmap']
        if not isinstance(custom_cmap, list):
            custom_cmap = [kwargs['cmap']]*num_plots
        else:
            if not len(custom_cmap) == num_plots:
                raise RuntimeError('Cmap list does not have enough items')
    else:
        custom_cmap = ['viridis']*num_plots

    if 'log' in kwargs and kwargs['log'] is True:
        logscale = kwargs['log']
    else:
        logscale = False

    if 'colorb' in kwargs:
        colorb = kwargs['colorb']
    else:
        colorb = True
    
    if 'dpi' in kwargs:
        dpi = kwargs['dpi']
    else:
        dpi = 90

    if 'title' in kwargs:
        tit = kwargs['title']
    else:
        tit = None
    
    if 'horsp' in kwargs:
        hor_spacing = kwargs['horsp']
    else:
        hor_spacing = 0.4
    
    if 'versp' in kwargs:
        ver_spacing = kwargs['versp']
    else:
        ver_spacing = 0.2

    # --------------------------------------------------------------------------

    subplot_size = 4
    if rows == 0:
        raise ValueError('Rows must be a positive integer')
    fig = figure(figsize=(cols * subplot_size, rows * subplot_size), dpi=dpi)
    
    if tit is not None:
        fig.suptitle(tit, fontsize=14)
    
    for i, v in enumerate(range(num_plots)):
        image = data[i].copy()
        frame_size = image.shape[0]  # assuming square frames
        cy, cx = frame_center(image)
        if grid_spacing is None:
            if cy < 10:
                grid_spacing = 1
            elif cy >= 10:
                if cy % 2 == 0:
                    grid_spacing = 4
                else:
                    grid_spacing = 5

        v += 1
        ax = subplot(rows, cols, v)
        ax.set_aspect('equal')
        if logscale:
            image += np.abs(image.min())
            norm = colors.LogNorm(vmin=max(image.min(), 0.1), vmax=image.max())
        else:
            norm = None
        im = ax.imshow(image, cmap=custom_cmap[i], interpolation='nearest',
                       origin='lower', vmin=vmin[i], vmax=vmax[i],
                       norm=norm)

        if show_circle:
            for j in range(len(coor_circle)):
                x = coor_circle[j][0]
                y = coor_circle[j][1]
                circle = Circle((x, y), radius=circle_rad, color='white',
                                fill=False, alpha=circle_alpha[j])
                ax.add_artist(circle)
                if circle_label:
                    cirlabel = str(int(x))+','+str(int(y))
                    ax.text(x, y+1.8*circle_rad, cirlabel, fontsize=8,
                            color='white', family='monospace', ha='center',
                            va='top', weight='bold', alpha=circle_alpha[j])

        if show_cross:
            ax.scatter([coor_cross[0]], [coor_cross[1]], marker='+',
                       color='white', alpha=cross_alpha)

        if show_center:
            ax.axhline(cx, xmin=0, xmax=frame_size, alpha=0.5,
                       linestyle='dashed', color='white', lw=0.8)
            ax.axvline(cy, ymin=0, ymax=frame_size, alpha=0.5,
                       linestyle='dashed', color='white', lw=0.8)

        if show_arrow:
            ax.arrow(coor_arrow[0]+arrow_length+arrow_shiftx, coor_arrow[1],
                     -arrow_length, 0, color='white', head_width=10,
                     head_length=8, width=3, length_includes_head=True,
                     alpha=arrow_alpha)
            
        if label is not None:
            ax.annotate(label[i], xy=(labelpad,labelpad), color='white',
                        xycoords='axes pixels', weight='bold', size=labelsize)    

        if colorb:
            # create an axes on the right side of ax. The width of cax is 5%
            # of ax and the padding between cax and ax wis fixed at 0.05 inch
            divider = make_axes_locatable(ax)
            cax = divider.append_axes("right", size="5%", pad=0.05)
            cb = colorbar(im, ax=ax, cax=cax, drawedges=False)
            cb.outline.set_linewidth(0.1)
            cb.ax.tick_params(labelsize=8) 

        if grid:
            ax.tick_params(axis='both', which='minor')
            minor_ticks = np.arange(0, data[i].shape[0], grid_spacing)
            ax.set_xticks(minor_ticks, minor=True)
            ax.set_yticks(minor_ticks, minor=True)
            ax.grid('on', which='minor', color=grid_color, linewidth=0.5,
                    alpha=grid_alpha, linestyle='dashed')
        else:
            ax.grid('off')

        if angscale:
            # Converting axes from pixels to arcseconds
            half_num_ticks = int(np.round(cy // angticksep))

            # Calculate the pixel locations at which to put ticks
            ticks = []
            for t in range(half_num_ticks, -half_num_ticks-1, -1):
                # Avoid ticks not showing on the last pixel
                if not cy - t * angticksep == frame_size:
                    ticks.append(cy - t * angticksep)
                else:
                    ticks.append((cy - t * angticksep) - 1)
            ax.set_xticks(ticks)
            ax.set_yticks(ticks)

            # Corresponding distance in arcseconds, measured from the center
            labels = []
            for t in range(half_num_ticks, -half_num_ticks-1, -1):
                labels.append(-t * (angticksep * pxscale))
            ax.set_xticklabels(labels)
            ax.set_yticklabels(labels)
            ax.set_xlabel("arcseconds", fontsize=12)
            ax.set_ylabel("arcseconds", fontsize=12)
            ax.tick_params(axis='both', which='major', labelsize=10)

        if not show_axis:
            ax.set_axis_off()
    
    fig.subplots_adjust(wspace=hor_spacing, hspace=ver_spacing)
    if save:
        savefig(savepath, dpi=dpi, bbox_inches='tight')
        close();
        if getfig:
            return fig
    else:
        show()
        if getfig:
            return fig


def plot_surface(image, center_xy=None, size=15, output=False, title=None,
                 zlim=None, **kwargs):
    """
    Create a surface plot from image.
    
    By default, the whole image is plotted. The 'center' and 'size' attributes 
    allow to crop the image.
        
    Parameters
    ----------
    image : numpy.array
        The image as a numpy.array.
    center_xy : tuple of 2 int or None, optional
        If None, the whole image will be plotted. Otherwise, it grabs a square
        subimage at the position X,Y from the image.
    size : int (optional, default=15)
        It corresponds to the size of a square in the image.
    output : {False, True}, bool optional
        Whether to output the grids and intensities or not.
    title : str, optional
        Sets the title for the surface plot.
    zlim : tuple, optional
        Tuple with the range of values (min,max) for the Z axis.
    kwargs:
        Additional attributes are passed to the matplotlib figure() and 
        plot_surface() method.        
    
    Returns
    -------
    out : tuple of 3 numpy.array
        x and y for the grid, and the intensity
        
    """
    if not isinstance(center_xy, tuple):
        raise ValueError('`center_xy` must be a tuple')

    if center_xy is not None:
        cx = int(center_xy[0])
        cy = int(center_xy[1])
        if size % 2:            # if size is odd
            x = np.outer(np.arange(0, size, 1), np.ones(size))
        else:                   # otherwise, size is even
            x = np.outer(np.arange(0, size+1, 1), np.ones(size+1))
        y = x.copy().T            
        z = image[cy-size//2: cy+size//2+1, cx-size//2: cx+size//2+1]
    else:
        cy, cx = frame_center(image)
        if size is not None:
            if size % 2:
                x = np.outer(np.arange(0,size,1), np.ones(size))
            else: 
                x = np.outer(np.arange(0,size+1,1), np.ones(size+1))
            y = x.copy().T
            z = image[cy-size//2:cy+size//2+1,cx-size//2:cx+size//2+1]
        else:
            size = image.shape[0]
            x = np.outer(np.arange(0, size, 1), np.ones(size))
            y = x.copy().T 
            z = image        
    
    figure(figsize=kwargs.pop('figsize', (6, 6)))
    ax = axes(projection='3d')
    ax.dist = 10
    ax.plot_surface(x, y, z, rstride=1, cstride=1, linewidth=0, **kwargs) 
    ax.set_xlabel('x')
    ax.set_ylabel('y')
    ax.set_zlabel('I(x,y)')
    if zlim is not None:
        if isinstance(zlim, tuple):
            ax.set_zlim(zlim[0], zlim[1])
    if title is not None:
        ax.set_title(title)
    show()
    
    if output:
        return x, y, z
 

def lines_of_code():
    """ Calculates the lines of code for VIP pipeline. Not objective measure 
    of developer's work! (note to self). 
    """
    cur_path = os.path.dirname(os.path.abspath(__file__))
    path = cur_path[:-len('var')]

    ignore_set = set(["__init__.py"])
    
    loclist = []

    for pydir, _, pyfiles in os.walk(path):
        if 'exlib/' not in pydir:
            for pyfile in pyfiles:
                if pyfile not in ignore_set and pyfile.endswith(".py"):
                    totalpath = os.path.join(pydir, pyfile)
                    loclist.append((len(open(totalpath,"r").read().splitlines()),
                                    totalpath.split(path)[1]))

    for linenumbercount, filename in loclist: 
        print("{:05d} lines in {}".format(linenumbercount, filename))

    msg = "\nTotal: {} lines in ({}) excluding external libraries."
    print(msg.format(sum([x[0] for x in loclist]), path))


def get_fwhm(lambd, diameter, pxscale):
    """ Returns the instrument FWHM [px] given the wavelenght [m], diameter [m]
    and plate/pixel scale [arcs/px]. In vip_hci/conf/param.py can be found
    dictionaries with the parameters for different instruments.
    """
    fwhm = lambd/diameter*206265/pxscale
    return fwhm

 



    



#! /usr/bin/env python

"""
Module with HCICube class.
"""

from __future__ import division, print_function

__author__ = 'Carlos Alberto Gomez Gonzalez'
__all__ = ['HCICube',
           'HCIFrame']

import numpy as np
from .fits import open_fits, write_fits, append_extension
from .preproc import frame_crop, frame_px_resampling, frame_rotate, frame_shift
from .preproc import (cube_collapse, cube_crop_frames, cube_derotate,
                      cube_drop_frames, cube_detect_badfr_correlation,
                      cube_detect_badfr_pxstats, cube_px_resampling,
                      cube_subsample, cube_recenter_2dfit,
                      cube_recenter_dft_upsampling)
from .var import frame_filter_lowpass, frame_filter_highpass, frame_center
from .var import (cube_filter_highpass, cube_filter_lowpass, mask_circle,
                  pp_subplots)
from .stats import frame_basic_stats, frame_histo_stats
from .stats import cube_basic_stats, cube_distance
from .phot import (frame_quick_report, cube_inject_companions, snr_ss,
                   snr_peakstddev, snrmap, snrmap_fast, detection,
                   normalize_psf)
from .conf.utils_conf import check_array


class HCIFrame:
    """ High-contrast imaging frame (2d array).

    Parameters
    ----------
    image : array_like
        2d array.
    hdu : int, optional
        If ``cube`` is a String, ``hdu`` indicates the HDU from the FITS file.
        By default the first HDU is used.
    fwhm : float, optional
        The FWHM associated with this dataset (instrument dependent). Required
        for several methods (operations on the cube).
    """
    def __init__(self, image, hdu=0, fwhm=None):
        """ HCIFrame object initialization. """
        if isinstance(image, str):
            self.image = open_fits(image, hdu, verbose=False)
        elif isinstance(image, np.ndarray):
            if not image.ndim == 2:
                raise ValueError('`Image` array has wrong dimensions')
            self.image = image
        else:
            raise ValueError('`Image` has a wrong type')
        print('Frame shape: {}'.format(self.image.shape))

        self.fwhm = fwhm
        if self.fwhm is not None:
            print('FWHM: {}'.format(self.fwhm))

    def crop(self, size, xy=None, force=False):
        """ Cropping the frame.
        """
        self.image = frame_crop(self.image, size, xy, force, verbose=True)

    def detect_blobs(self, psf, bkg_sigma=1, method='lpeaks',
                     matched_filter=False, mask=True, snr_thresh=5, plot=True,
                     debug=False, verbose=False, save_plot=None,
                     plot_title=None, angscale=False):
        """ Detecting blobs on the 2d array.
        """
        self.detection_results = detection(self.image, psf, bkg_sigma, method,
                                           matched_filter, mask, snr_thresh,
                                           plot, debug, True, verbose,
                                           save_plot, plot_title, angscale)

    def filter(self, method, mode, median_size=5, kernel_size=5, fwhm_size=5,
               btw_cutoff=0.2, btw_order=2, gauss_mode='conv'):
        """ High/low pass filtering the frames of the image.

        Parameters
        ----------
        method : {'lp', 'hp'}

        mode : {'median', 'gauss'}
        {'laplacian', 'laplacian-conv', 'median-subt', 'gauss-subt', 'fourier-butter'}
        """
        if method == 'hp':
            self.image = frame_filter_highpass(self.image, mode, median_size,
                                               kernel_size, fwhm_size,
                                               btw_cutoff, btw_order)
        elif method == 'lp':
            self.image = frame_filter_lowpass(self.image, mode, median_size,
                                              fwhm_size, gauss_mode)
        else:
            raise ValueError('Filtering mode not recognized')
        print('Image successfully filtered')

    def get_center(self, verbose=True):
        """ Getting the coordinates of the center of the image.
        """
        cent = frame_center(self.image, verbose)
        return cent

    def plot(self, **kwargs):
        """ Plotting the 2d array.

        Parameters in **kwargs
        ----------------------
        angscale : bool
            If True, the axes are displayed in angular scale (arcsecs).
        angticksep : int
            Separation for the ticks when using axis in angular scale.
        arrow : bool
            To show an arrow pointing to input px coordinates.
        arrowalpha : float
            Alpha transparency for the arrow.
        arrowlength : int
            Length of the arrow, 20 px by default.
        arrowshiftx : int
            Shift in x of the arrow pointing position, 5 px by default.
        axis : bool
            Show the axis, on by default.
        circle : list of tuples
            To show a circle at given px coordinates, list of tuples.
        circlerad : int
            Radius of the circle, 6 px by default.
        cmap : str
            Colormap to be used, 'viridis' by default.
        colorb : bool
            To attach a colorbar, on by default.
        cross : tuple of float
            If provided, a crosshair is displayed at given px coordinates.
        crossalpha : float
            Alpha transparency of thr crosshair.
        dpi : int
            Dots per inch, for plot quality.
        getfig : bool
            Returns the matplotlib figure.
        grid : bool
            If True, a grid is displayed over the image, off by default.
        gridalpha : float
            Alpha transparency of the grid.
        gridcolor : str
            Color of the grid lines.
        gridspacing : int
            Separation of the grid lines in pixels.
        horsp : float
            Horizontal gap between subplots.
        label : str or list of str
            Text for annotating on subplots.
        labelpad : int
            Padding of the label from the left bottom corner.
        labelsize : int
            Size of the labels.
        log : bool
            Log colorscale.
        maxplots : int
            When the input (*args) is a 3d array, maxplots sets the number of
            cube slices to be displayed.
        pxscale : float
            Pixel scale in arcseconds/px. Default 0.01 for Keck/NIRC2.
        rows : int
            How many rows (subplots in a grid).
        save : str
            If a string is provided the plot is saved using this as the path.
        showcent : bool
            To show a big crosshair at the center of the frame.
        title : str
            Title of the plot(s), None by default.
        vmax : int
            For stretching the displayed pixels values.
        vmin : int
            For stretching the displayed pixels values.
        versp : float
            Vertical gap between subplots.
        """
        pp_subplots(self.image, **kwargs)

    def rescale(self, scale, imlib='ndimage', interpolation='bicubic',
                verbose=True):
        """ Resampling the image (upscaling or downscaling).
        """
        self.image = frame_px_resampling(self.image, scale, imlib, interpolation,
                                         verbose)

    def rotate(self, angle, imlib='opencv', interpolation='lanczos4', cxy=None):
        """ Rotating the image.
        """
        self.image = frame_rotate(self.image, angle, imlib, interpolation, cxy)
        print('Image successfully rotated')

    def save(self, path):
        """ Writing to FITS file.
        """
        write_fits(path, self.image)

    def shift(self, shift_y, shift_x, imlib='opencv', interpolation='lanczos4'):
        """ Shifting the image.
        """
        self.image = frame_shift(self.image, shift_y, shift_x, imlib,
                                 interpolation)
        print('Image successfully shifted')

    def snr(self, source_xy, method='student', plot=False, verbose=True):
        """ Calculating the S/N for a test resolution element ``source_xy``.

        Parameters
        ----------
        source_xy : tuple of floats
            X and Y coordinates of the planet or test speckle.
        method : {'student', 'classic'}, str optional
            With 'student' the small sample statistics (Mawet et al. 2014) is
            used. With 'classic', the S/N is estimated with the old approach
            using the standard deviation of independent pixels.
        plot : bool, optional
            Plots the frame and the apertures considered for clarity.
        verbose : bool, optional
            Chooses whether to print some output or not.

        Returns
        -------
        snr_val : float
            Value of the S/N for ``source_xy``.
        """
        if self.fwhm is None:
            raise ValueError('FWHM has not been set')

        if method == 'student':
            snr_val = snr_ss(self.image, source_xy, self.fwhm, False, plot,
                             verbose)
        elif method == 'classic':
            snr_val = snr_peakstddev(self.image, source_xy, self.fwhm, False,
                                     plot, verbose)
        else:
            raise ValueError('S/N estimation method not recognized')
        return snr_val

    def snrmap(self, method='student', approx=False, plot=True,
               source_mask=None, nproc=None, verbose=True):
        """ Generating the S/N map for the image.

        Parameters
        ----------
        method : {'student', 'classic'}, str optional
            With 'student' the small sample statistics (Mawet et al. 2014) is
            used. With 'classic', the S/N is estimated with the old approach
            using the standard deviation of independent pixels.
        approx : bool, optional
            If True, the function ``vip_hci.phot.snrmap_fast`` is used instead
            of ``vip_hci.phot.snrmap``.
        plot : bool, optional
            If True plots the S/N map. True by default.
        source_mask : array_like, optional
            If exists, it takes into account existing sources. The mask is a
            ones 2d array, with the same size as the input frame. The centers
            of the known sources have a zero value.
        nproc : int or None
            Number of processes for parallel computing.

        Returns
        -------
        map : HCIFrame object
            S/N map.
        """
        if self.fwhm is None:
            raise ValueError('FWHM has not been set')

        if approx:
            map = snrmap_fast(self.image, self.fwhm, nproc, plot, verbose)
        else:
            if method == 'student':
                mode = 'sss'
            elif method == 'classic':
                mode = 'peakstddev'
            map = snrmap(self.image, self.fwhm, plot, mode, source_mask, nproc,
                         verbose=verbose)
        return HCIFrame(map)

    def stats(self, region='circle', radius=5, xy=None, annulus_inner_radius=0,
              annulus_width=5, source_xy=None, verbose=True, plot=True):
        """ Calculating statistics on the image, both in the full-frame and in
        a region (circular aperture or annulus). Also, the S/N of the either
        ``source_xy`` or the max pixel is calculated.

        Parameters
        ----------
        region : {'circle', 'annulus'}, str optional
            Region in which basic statistics (mean, stddev, median and max) are
            calculated.
        radius : int, optional
            Radius of the circular aperture.
        xy : tuple of floats, optional
            Center of the circular aperture.
        annulus_inner_radius : int, optional
            Inner radius of the annular region.
        annulus_width : int, optional
            Width of the annular region.
        source_xy : tuple of floats, optional
            Coordinates for which the S/N information will be obtained. If None,
            the S/N is estimated for the pixel with the maximum value.
        verbose : bool, optional
            Whether to print out the values of the calculated statistics.
        plot : bool, optional
            Whether to plot the frame, histograms and region.
        """
        res_region = frame_basic_stats(self.image, region, radius, xy,
                                       annulus_inner_radius, annulus_width,
                                       plot, True)
        if verbose:
            if region == 'circle':
                msg = 'Stats in circular aperture of radius: {}pxs'
                print(msg.format(radius))
            elif region == 'annulus':
                msg = 'Stats in annulus. Inner_rad: {}pxs, width: {}pxs'
                print(msg.format(annulus_inner_radius, annulus_width))
            mean, std_dev, median, maxi = res_region
            msg = 'Mean: {:.3f}, Stddev: {:.3f}, Median: {:.3f}, Max: {:.3f}'
            print(msg.format(mean, std_dev, median, maxi))

        res_ff = frame_histo_stats(self.image, plot)
        if verbose:
            mean, median, std, maxim, minim = res_ff
            print('Stats in the whole frame:')
            msg = 'Mean: {:.3f}, Stddev: {:.3f}, Median: {:.3f}, Max: {:.3f}, '
            msg += 'Min: {:.3f}'
            print(msg.format(mean, std, median, maxim, minim))

        print('\nS/N info:')
        _ = frame_quick_report(self.image, self.fwhm, source_xy, verbose)


class HCICube:
    """ High-contrast imaging sequence/cube class.

    Parameters
    ----------
    cube : str or numpy array
        3d or 4d high-contrast image sequence. If a string is provided, cube is
        interpreted as the path of the FITS file containing the sequence.
    hdu : int, optional
        If ``cube`` is a String, ``hdu`` indicates the HDU from the FITS file.
        By default the first HDU is used.
    angles : list or numpy array, optional
        The vector of parallactic angles.
    wavelengths : list or numpy array, optional
        The vector of wavelengths (to be used as scaling factors).
    fwhm : float, optional
        The FWHM associated with this dataset (instrument dependent). Required
        for several methods (operations on the cube).
    px_scale : float, optional
        The pixel scale associated with this dataset (instrument dependent).
    psf : numpy array, optional
        The PSF template associated with this dataset.
    psfn : numpy array, optional
        Normalized/cropped/centered version of the PSF template associated with
        this dataset.
    cuberef : str or numpy array
        3d or 4d high-contrast image sequence. To be used as a reference cube.
    """
    def __init__(self, cube, hdu=0, angles=None, wavelengths=None, fwhm=None,
                 px_scale=None, psf=None, psfn=None, cuberef=None):
        """ Initialization of the HCICube object.
        """
        # Loading the 3d/4d cube or image sequence
        if isinstance(cube, str):
            self.cube = open_fits(cube, hdu, verbose=False)
        elif isinstance(cube, np.ndarray):
            if not (cube.ndim == 3 or cube.ndim == 4):
                raise ValueError('`Cube` array has wrong dimensions')
            self.cube = cube
        else:
            raise TypeError('`Cube` has a wrong type')
        print('Cube array shape: {}'.format(self.cube.shape))
        if self.cube.ndim == 3:
            self.n, self.y, self.x = self.cube.shape
        elif self.cube.ndim == 4:
            self.w, self.n, self.y, self.x = self.cube.shape

        # Loading the reference cube
        if cuberef is not None:
            if isinstance(cuberef, str):
                self.cuberef = open_fits(cuberef, hdu, verbose=False)
            elif isinstance(cuberef, np.ndarray):
                msg = '`Cuberef` array has wrong dimensions'
                if not cuberef.ndim == 3:
                    raise ValueError(msg)
                if not cuberef.shape[1] == self.y:
                    raise ValueError(msg)
                self.cuberef = cuberef
            elif isinstance(cuberef, HCICube):
                msg = '`Cuberef` array has wrong dimensions'
                if not cuberef.cube.ndim == 3:
                    raise ValueError(msg)
                if not cuberef.cube.shape[1] == self.y:
                    raise ValueError(msg)
                self.cuberef = cuberef.cube
            else:
                raise TypeError('`Cuberef` has a wrong type')
            print('Cuberef array shape: {}'.format(self.cuberef.shape))

        # Loading the angles (ADI)
        if isinstance(angles, str):
            self.angles = open_fits(angles, verbose=False)
        else:
            self.angles = angles
        if self.angles is not None:
            print('Angles array shape: {}'.format(self.angles.shape))
            # Checking the shape of the angles vector
            self.angles = check_array(self.angles, dim=1,
                                      name='Parallactic angles vector')
            if not self.angles.shape[0] == self.n:
                raise ValueError('Parallactic angles vector has a wrong shape')

        # Loading the scaling factors (mSDI)
        if isinstance(wavelengths, str):
            self.wavelengths = open_fits(wavelengths, verbose=False)
        else:
            self.wavelengths = wavelengths
        if self.wavelengths is not None:
            print('Wavelengths array shape: {}'.format(self.wavelengths.shape))
            # Checking the shape of the scaling vector
            self.wavelengths = check_array(self.wavelengths, dim=1,
                                           name='Wavelengths vector')
            if not self.wavelengths.shape[0] == self.w:
                raise ValueError('Wavelengths vector has a wrong shape')

        # Loading the PSF
        if isinstance(psf, str):
            self.psf = open_fits(psf, verbose=False)
        else:
            self.psf = psf
        if self.psf is not None:
            print('PSF array shape: {}'.format(self.psf.shape))
            # Checking the shape of the PSF array
            if not self.psf.ndim == self.cube.ndim - 1:
                msg = 'PSF array has a wrong shape. Must have {} dimensions, '
                msg += 'got {} instead'
                raise ValueError(msg.format(self.cube.ndim - 1, self.psf.ndim))

        # Loading the normalized PSF
        if isinstance(psfn, str):
            self.psfn = open_fits(psfn, verbose=False)
        else:
            self.psfn = psfn
        if self.psfn is not None:
            print('Normalized PSF array shape: {}'.format(self.psfn.shape))
            # Checking the shape of the PSF array
            if not self.psfn.ndim == self.cube.ndim - 1:
                msg = 'Normalized PSF array has a wrong shape. Must have {} '
                msg += 'dimensions, got {} instead'
                raise ValueError(msg.format(self.cube.ndim - 1, self.psfn.ndim))

        self.fwhm = fwhm
        if self.fwhm is not None:
            if self.cube.ndim == 4:
                self.fwhm = check_array(self.fwhm, 1, 'FHWM')
            elif self.cube.ndim == 3:
                print('FWHM: {}'.format(self.fwhm))
        self.px_scale = px_scale
        if self.px_scale is not None:
            print('Pixel/plate scale: {}'.format(self.px_scale))

    def collapse(self, mode='median', n=50):
        """ Collapsing the sequence into a 2d array.

        # TODO: support 4d case.
        """
        frame = cube_collapse(self.cube, mode, n)
        print('Cube successfully collapsed')
        return HCIFrame(frame)

    def crop_frames(self, size, xy=None, force=False):
        """ Cropping the frames of the sequence (3d or 4d cube).

        Parameters
        ----------
        size : int
            New size of the (square) frames.
        xy : tuple of ints
            X, Y coordinates of new frame center. If you are getting the
            coordinates from ds9 subtract 1, python has 0-based indexing.
        force : bool, optional
            ``Size`` and the original size of the frames must be both even or odd.
            With ``force`` set to True this condition can be avoided.
        """
        self.cube = cube_crop_frames(self.cube, size, xy, force, verbose=True)

    def derotate(self, imlib='opencv', interpolation='lanczos4', cxy=None,
                 nproc=1):
        """ Derotating the frames of the sequence according to the parallactic
        angles.
        """
        if self.angles is None:
            raise ValueError('Parallactic angles vector has not been set')

        self.cube = cube_derotate(self.cube, self.angles, imlib,
                                  interpolation, cxy, nproc)
        print('Cube successfully derotated')

    def drop_frames(self, n, m):
        """ Slicing the cube using the `n` (initial) and `m` (final) indices in
        a 1-indexed fashion.

        # TODO: support 4d case.
        """
        self.cube = cube_drop_frames(self.cube, n, m, self.angles)

    def filter(self, method, mode, median_size=5, kernel_size=5, fwhm_size=5,
               btw_cutoff=0.2, btw_order=2, gauss_mode='conv', verbose=True):
        """ High/low pass filtering the frames of the cube.

        # TODO: support 4d case.

        Parameters
        ----------
        method : {'lp', 'hp'}

        mode : {'median', 'gauss'}
        {'laplacian', 'laplacian-conv', 'median-subt', 'gauss-subt', 'fourier-butter'}
        """
        if method == 'hp':
            self.cube = cube_filter_highpass(self.cube, mode, median_size,
                                              kernel_size, fwhm_size,
                                              btw_cutoff, btw_order, verbose)
        elif method == 'lp':
            self.cube = cube_filter_lowpass(self.cube, mode, median_size,
                                             fwhm_size, gauss_mode, verbose)
        else:
            raise ValueError('Filtering mode not recognized')

    def frame_distances(self, frame, region='full', dist='sad',
                        inner_radius=None, width=None, plot=True):
        """ Calculating the frame distance/correlation with respect to a
        reference image.

        # TODO: support 4d case.
        """
        _ = cube_distance(self.cube, frame, region, dist, inner_radius, width,
                          plot)

    # TODO: support 4d case.
    def frame_stats(self, region='circle', radius=5, xy=None,
                    annulus_inner_radius=0, annulus_width=5, plot=False):
        """ Calculating statistics on a ``region`` (circular aperture or
        annulus) of each image of the sequence.

        Parameters
        ----------
        region : {'circle', 'annulus'}, str optional
            Region in which basic statistics (mean, stddev, median and max) are
            calculated.
        radius : int, optional
            Radius of the circular aperture.
        xy : tuple of floats, optional
            Center of the circular aperture.
        annulus_inner_radius : int, optional
            Inner radius of the annular region.
        annulus_width : int, optional
            Width of the annular region.
        plot : bool, optional
            Whether to plot the frame, histograms and region.
        """
        _ = cube_basic_stats(self.cube, region, radius, xy,
                             annulus_inner_radius, annulus_width, plot, False)

    def inject_companions(self, flux, rad_dists, n_branches=1,
                          theta=0, imlib='opencv', interpolation='lanczos4',
                          verbose=True):
        """ Injection of fake companions in 3d or 4d cubes.

        # TODO: support the injection of a Gaussian/Moffat kernel.
        # TODO: return array/HCICube object instead?
        """
        if self.psf is None:
            raise ValueError('PSf array has not been set')
        if self.px_scale is None:
            raise ValueError('Pixel/plate scale has not been set')

        self.cube = cube_inject_companions(self.cube, self.psf, self.angles,
                                            flux, self.px_scale, rad_dists,
                                            n_branches, theta, imlib,
                                            interpolation, verbose)

    def load_angles(self, angles, hdu):
        """ Loads the PA vector from a FITS file. It is possible to specify the
        HDU.
        """
        if isinstance(angles, str):
            self.angles = open_fits(angles, hdu)
        elif isinstance(angles, (list, np.ndarray)):
            self.angles = angles
        else:
            msg = '`Angles` has a wrong type. Must be a list or 1d np.ndarray'
            raise ValueError(msg)

    def load_wavelengths(self, wavelengths, hdu):
        """ Loads the scaling factors vector from a FITS file. It is possible to
        specify the HDU.

        Parameters
        ----------
        wavelengths :
        """
        if isinstance(wavelengths, str):
            self.wavelengths = open_fits(wavelengths, hdu)
        elif isinstance(wavelengths, (list, np.ndarray)):
            self.wavelengths = wavelengths
        else:
            msg = '`wavelengths` has a wrong type. Must be a list or np.ndarray'
            raise ValueError(msg)

    def mask_center(self, radius, fillwith=0, mode='in'):
        """ Masking the values inside/outside a centered circular aperture.

        # TODO: support 4d case.
        """
        self.cube = mask_circle(self.cube, radius, fillwith, mode)

    def normalize_psf(self, fwhm='fit', size=None, threshold=None,
                      mask_core=None, model='gauss', imlib='opencv',
                      interpolation='lanczos4', force_odd=True, verbose=True):
        """ Normalizes a PSF (2d or 3d array), to have the flux in a 1xFWHM
        aperture equal to one. It also allows to crop the array and center the
        PSF at the center of the frame(s).
        """
        if self.psf is None:
            raise ValueError('PSF array has not been loaded')
        self.psfn = normalize_psf(self.psf, fwhm, size, threshold, mask_core,
                                  model, imlib, interpolation, force_odd,
                                  False, verbose)
        print('Normalized PSF array shape: {}'.format(self.psfn.shape))

    def plot(self, **kwargs):
        """ Plotting ``maxplots`` frames of the cube (3D case).

        Parameters in **kwargs
        ----------------------
        angscale : bool
            If True, the axes are displayed in angular scale (arcsecs).
        angticksep : int
            Separation for the ticks when using axis in angular scale.
        arrow : bool
            To show an arrow pointing to input px coordinates.
        arrowalpha : float
            Alpha transparency for the arrow.
        arrowlength : int
            Length of the arrow, 20 px by default.
        arrowshiftx : int
            Shift in x of the arrow pointing position, 5 px by default.
        axis : bool
            Show the axis, on by default.
        circle : list of tuples
            To show a circle at given px coordinates, list of tuples.
        circlerad : int
            Radius of the circle, 6 px by default.
        cmap : str
            Colormap to be used, 'viridis' by default.
        colorb : bool
            To attach a colorbar, on by default.
        cross : tuple of float
            If provided, a crosshair is displayed at given px coordinates.
        crossalpha : float
            Alpha transparency of thr crosshair.
        dpi : int
            Dots per inch, for plot quality.
        getfig : bool
            Returns the matplotlib figure.
        grid : bool
            If True, a grid is displayed over the image, off by default.
        gridalpha : float
            Alpha transparency of the grid.
        gridcolor : str
            Color of the grid lines.
        gridspacing : int
            Separation of the grid lines in pixels.
        horsp : float
            Horizontal gap between subplots.
        label : str or list of str
            Text for annotating on subplots.
        labelpad : int
            Padding of the label from the left bottom corner.
        labelsize : int
            Size of the labels.
        log : bool
            Log colorscale.
        maxplots : int
            When the input (*args) is a 3d array, maxplots sets the number of
            cube slices to be displayed.
        pxscale : float
            Pixel scale in arcseconds/px. Default 0.01 for Keck/NIRC2.
        rows : int
            How many rows (subplots in a grid).
        save : str
            If a string is provided the plot is saved using this as the path.
        showcent : bool
            To show a big crosshair at the center of the frame.
        title : str
            Title of the plot(s), None by default.
        vmax : int
            For stretching the displayed pixels values.
        vmin : int
            For stretching the displayed pixels values.
        versp : float
            Vertical gap between subplots.
        """
        # TODO: argument to slice the array?

        if self.cube.ndim == 3:
            pp_subplots(self.cube, **kwargs)
        elif self.cube.ndim == 4:
            for i in range(0, self.cube.shape[0], 10):
                tits = 'Wavelength '+str(i + 1)+', first 5 frames'
                pp_subplots(self.cube[i], title=tits,
                            maxplots=5, **kwargs)

    def recenter(self, method='2dfit', xy=None, subi_size=5, model='gauss',
                 nproc=1, imlib='opencv', interpolation='lanczos4',
                 offset=None, negative=False, threshold=False,
                 save_shifts=False, cy_1=None, cx_1=None, upsample_factor=100,
                 verbose=True, debug=False, plot=True):
        """ Frame to frame recentering.
        """
        if self.fwhm is None:
            raise ValueError('FWHM has not been set')

        if method == '2d_fitting':
            self.cube = cube_recenter_2dfit(self.cube, xy, self.fwhm,
                                subi_size, model, nproc, imlib, interpolation,
                                offset, negative, threshold, save_shifts, False,
                                verbose, debug, plot)
        elif method == 'dft_upsampling':
            self.cube = cube_recenter_dft_upsampling(self.cube, cy_1, cx_1,
                                negative, self.fwhm, subi_size, upsample_factor,
                                imlib, interpolation, False, verbose,
                                save_shifts, debug)
        else:
            # TODO support other recentering methods from vip_hci.preproc
            raise ValueError('Method not recognized')

    def remove_badframes(self, method='corr', frame_ref=None, crop_size=30,
                         dist='pearson', percentile=20, stat_region='annulus',
                         inner_radius=10, width=10, top_sigma=1.0,
                         low_sigma=1.0, window=None, plot=True, verbose=True):
        """ Finding outlying/bad frames and slicing the cube accordingly.

        Parameters
        ----------
        method : {'corr', 'pxstats'}, str optional

        """
        if method == 'corr':
            if frame_ref is None:
                print("Correlation method selected but `frame_ref` is missing")
                print("Setting the 1st frame as the reference")
                frame_ref = 0

            self.good_indices, _ = cube_detect_badfr_correlation(self.cube,
                                            frame_ref, crop_size, dist,
                                            percentile, plot, verbose)
        elif method == 'pxstats':
            self.good_indices, _ = cube_detect_badfr_pxstats(self.cube,
                                            stat_region, inner_radius, width,
                                            top_sigma, low_sigma, window, plot,
                                            verbose)
        else:
            raise ValueError('Bad frames detection method not recognized')

        self.cube = self.cube[self.good_indices]
        print("New cube shape: {}".format(self.cube.shape))
        if self.angles is not None:
            self.angles = self.angles[self.good_indices]
            msg = "New parallactic angles vector shape: {}"
            print(msg.format(self.angles.shape))

    def rescale(self, scale, imlib='ndimage', interpolation='bicubic',
                verbose=True):
        """ Resampling the pixels (upscaling or downscaling the frames).
        """
        self.cube = cube_px_resampling(self.cube, scale, imlib, interpolation,
                                        verbose)

    def save(self, path, dtype32=True):
        """ Writing to FITS file. If self.angles is present, then the angles
        are appended to the FITS file.

        Parameters
        ----------
        filename : string
            Full path of the fits file to be written.
        dtype32 : bool, optional
            If True the array is casted as a float32. When False, the array is
            usually saved in float64 precision.
        """
        write_fits(path, self.cube, dtype32)
        if self.angles is not None:
            append_extension(path, self.angles)

    def subsample(self, window, mode='mean'):
        """ Temporally sub-sampling the sequence (3d or 4d cube).

        Parameters
        ----------
        window : int
            Window for mean/median.
        mode : {'mean','median'}, optional
            Switch for choosing mean or median.
        """
        if self.angles is not None:
            self.cube, self.angles = cube_subsample(self.cube, window,
                                                     mode, self.angles)
        else:
            self.cube = cube_subsample(self.cube, window, mode)





#! /usr/bin/env python

"""
Module for image statistics.
"""

from __future__ import division, print_function

__author__ = 'Carlos Alberto Gomez Gonzalez'
__all__ = ['frame_histo_stats',
           'average_radial_profile']

import numpy as np
import pandas as pd
from matplotlib import pyplot as plt
from ..var import frame_center


def average_radial_profile(array, sep=1, collapse='median', plot=True):
    """ Calculates the average radial profile of an image. If a cube is
    passed it's first collapsed.

    Parameters
    ----------
    array : array_like, 2d or 3d
        Input image or cube.
    sep : int, optional
        The average radial profile is recorded every ``sep`` pixels.
    collapse : str, optional
        The method for combining the images, when array is cube.
    plot : bool, optional
        If True the profile is plotted.

    Returns
    -------
    df : dataframe
        Pandas dataframe with the radial profile and distances.

    """
    from ..preproc import cube_collapse
    if array.ndim == 3:
        frame = cube_collapse(array, mode=collapse)
    elif array.ndim == 2:
        frame = array

    cy, cx = frame_center(frame)
    x, y = np.indices((frame.shape))
    r = np.sqrt((x - cx) ** 2 + (y - cy) ** 2)
    r = r.astype(int)
    tbin = np.bincount(r.ravel(), frame.ravel())
    nr = np.bincount(r.ravel())
    radprofile = tbin / nr

    radists = np.arange(1, int(cy), sep)
    radprofile_radists = radprofile[radists - 1]
    nr_radists = nr[radists - 1]
    df = pd.DataFrame(
        {'rad': radists - 1, 'radprof': radprofile_radists, 'npx': nr_radists})

    if plot:
        plt.figure(figsize=(10, 4), dpi=100)
        plt.plot(radists - 1, radprofile_radists, '.-', alpha=0.6)
        plt.grid(which='both', alpha=0.6)
        plt.xlabel('Pixels')
        plt.ylabel('Counts')
        plt.minorticks_on()
        plt.xlim(0)

    return df


def frame_histo_stats(image_array, plot=True):
    """Plots a frame with a colorbar, its histogram and some statistics: mean,
    median, maximum, minimum and standard deviation values.  
    
    Parameters
    ----------
    image_array : array_like
        The input frame.  
    plot : bool, optional
        If True plots the frame and the histogram with the values.
        
    Return
    ------
    mean : float
        Mean value of array.
    median : float
        Median value of array.
    std : float
        Standard deviation of array.
    maxim : int or float
        Maximum value.
    minim : int or float
        Minimum value.
        
    """
    vector = image_array.flatten()
    mean = vector.mean()
    median = np.median(vector)
    maxim = vector.max()
    minim = vector.min()
    std = vector.std()
   
    if plot:
        fig, axes = plt.subplots(nrows=1, ncols=2, figsize=(10, 3))
        ax0, ax1 = axes.flat
        bins = int(np.sqrt(vector.shape[0]))
        txt = 'Mean = {:.3f}\n'.format(mean) + \
              'Median = {:.3f}\n'.format(median) +\
              'Stddev = {:.3f}\n'.format(std) +\
              'Max = {:.3f}\n'.format(maxim) +\
              'Min = {:.3f}\n\n'.format(minim)

        ax0.imshow(image_array, interpolation="nearest", origin ="lower",
                   cmap='viridis')
        ax0.set_title('Frame')
        ax0.grid('off')

        ax1.hist(vector, bins=bins, label=txt, alpha=0.5, histtype='stepfilled')
        ax1.set_yscale('log')
        ax1.set_title('Histogram')
        ax1.text(0.98, 0.98, txt, transform=ax1.transAxes, fontsize=12,
                 verticalalignment='top', horizontalalignment='right')
        plt.show()
        
    return mean, median, std, maxim, minim




#! /usr/bin/env python

"""
Full-frame PCA algorithm for ADI, ADI+RDI and ADI+mSDI (IFS data) cubes.
"""

from __future__ import division, print_function

__author__ = 'Carlos Alberto Gomez Gonzalez'
__all__ = ['pca',
           'pca_incremental',
           'pca_optimize_snr']

import numpy as np
import pandas as pd
import pyprind
from skimage import draw
from astropy.io import fits
from matplotlib import pyplot as plt
from sklearn.decomposition import IncrementalPCA
from .svd import svd_wrapper
from ..madi.adi_utils import _find_indices, _compute_pa_thresh
from .utils_pca import pca_annulus, scale_cube_for_pca as scpca
from ..preproc import (cube_derotate, cube_collapse, check_pa_vector,
                       check_scal_vector, cube_crop_frames)
from ..conf import timing, time_ini, check_enough_memory, get_available_memory
from ..var import frame_center, dist, prepare_matrix, reshape_matrix
from ..stats import descriptive_stats
from .. import phot

import warnings
warnings.filterwarnings("ignore", category=Warning)


def pca(cube, angle_list, cube_ref=None, scale_list=None, ncomp=1, ncomp2=1,
        svd_mode='lapack', scaling=None, adimsdi='double', mask_center_px=None,
        source_xy=None, delta_rot=1, fwhm=4, imlib='opencv',
        interpolation='lanczos4', collapse='median', check_mem=True,
        full_output=False, verbose=True, debug=False):
    """ Algorithm where the reference PSF and the quasi-static speckle pattern 
    are modeled using Principal Component Analysis. Depending on the input
    parameters this PCA function can work in ADI, RDI or SDI (IFS data) mode.
    
    ADI: If neither a reference cube nor a scaling vector are provided, the
    target cube itself is used to learn the PCs and to obtain a low-rank
    approximation model PSF (star + speckles).
    
    ADI + RDI: if a reference cube is provided (``cube_ref``), its PCs are used
    to reconstruct the target frames to obtain the model PSF (star + speckles).
    
    ADI + SDI (IFS data): if a scaling vector is provided (``scale_list``) and
    the cube is a 4d array [# channels, # adi-frames, Y, X], its assumed it 
    contains several multi-spectral ADI frames. A single or two stages PCA can
    be performed, depending on ``adimsdi``.
    
    Notes
    -----
    The full-frame ADI-PCA implementation is based on Soummer et al. 2012
    (http://arxiv.org/abs/1207.4197) and Amara & Quanz 2012
    (http://arxiv.org/abs/1207.6637).
    
    Parameters
    ----------
    cube : array_like, 3d or 4d
        Input cube (ADI or ADI+mSDI).
    angle_list : array_like, 1d
        Corresponding parallactic angle for each frame.    
    cube_ref : array_like, 3d, optional
        Reference library cube. For Reference Star Differential Imaging.
    scale_list : 
        Scaling factors in case of IFS data (ADI+mSDI cube). Usually, the
        scaling factors are the central channel wavelength divided by the
        shortest wavelength in the cube (more thorough approaches can be used
        to get the scaling factors). This scaling factors are used to re-scale
        the spectral channels and align the speckles.
    ncomp : int, optional
        How many PCs are used as a lower-dimensional subspace to project the
        target frames. For an ADI cube, ``ncomp`` is the number of PCs extracted
        from ``cube``. For the RDI case, when ``cube`` and ``cube_ref`` are
        provided, ``ncomp`` is the number of PCs obtained from ``cube_ref``.
        For an ADI+mSDI cube (e.g. SPHERE/IFS), if ``adimsdi`` is ``double``
        then ``ncomp`` is the number of PCs obtained from each multi-spectral
        frame (if ``ncomp`` is None then this stage will be skipped and the
        spectral channels will be combined without subtraction). If ``adimsdi``
        is ``single``, then ``ncomp`` is the number of PCs obtained from the
        whole set of frames (n_channels * n_adiframes).
    ncomp2 : int, optional
        Only used for ADI+mSDI cubes, when ``adimsdi`` is set to ``double``.
        ``ncomp2`` sets the number of PCs used in the second PCA stage (ADI
        fashion, using the residuals of the first stage). If None then the
        second PCA stage is skipped and the residuals are de-rotated and
        combined.
    svd_mode : {'lapack', 'arpack', 'eigen', 'randsvd', 'cupy', 'eigencupy', 'randcupy'}, str
        Switch for the SVD method/library to be used. ``lapack`` uses the LAPACK 
        linear algebra library through Numpy and it is the most conventional way 
        of computing the SVD (deterministic result computed on CPU). ``arpack`` 
        uses the ARPACK Fortran libraries accessible through Scipy (computation
        on CPU). ``eigen`` computes the singular vectors through the 
        eigendecomposition of the covariance M.M' (computation on CPU).
        ``randsvd`` uses the randomized_svd algorithm implemented in Sklearn 
        (computation on CPU). ``cupy`` uses the Cupy library for GPU computation
        of the SVD as in the LAPACK version. ``eigencupy`` offers the same 
        method as with the ``eigen`` option but on GPU (through Cupy). 
        ``randcupy`` is an adaptation of the randomized_svd algorithm, where all
        the computations are done on a GPU. 
    scaling : {None, 'temp-mean', 'spat-mean', 'temp-standard', 'spat-standard'}
        With None, no scaling is performed on the input data before SVD. With 
        "temp-mean" then temporal px-wise mean subtraction is done, with 
        "spat-mean" then the spatial mean is subtracted, with "temp-standard" 
        temporal mean centering plus scaling to unit variance is done and with
        "spat-standard" spatial mean centering plus scaling to unit variance is
        performed.  
    mask_center_px : None or int
        If None, no masking is done. If an integer > 1 then this value is the
        radius of the circular mask. 
    source_xy : tuple of int, optional 
        For ADI PCA, this triggers a frame rejection in the PCA library. 
        source_xy are the coordinates X,Y of the center of the annulus where the
        PA criterion will be used to reject frames from the library. 
    fwhm : float, optional
        Known size of the FHWM in pixels to be used. Default value is 4.
    imlib : str, optional
        See the documentation of the ``vip_hci.preproc.frame_rotate`` function.
    interpolation : str, optional
        See the documentation of the ``vip_hci.preproc.frame_rotate`` function.
    delta_rot : int, optional
        Factor for tunning the parallactic angle threshold, expressed in FWHM.
        Default is 1 (excludes 1xFHWM on each side of the considered frame).
    collapse : {'median', 'mean', 'sum', 'trimmean'}, str optional
        Sets the way of collapsing the frames for producing a final image.
    check_mem : bool, optional
        If True, it check that the input cube(s) are smaller than the available
        system memory.
    full_output: bool, optional
        Whether to return the final median combined image only or with other 
        intermediate arrays.  
    verbose : bool, optional
        If True prints intermediate info and timing. 
    debug : bool, optional
        Whether to print debug information or not.
    
    Returns
    -------
    frame : array_like, 2d    
        Median combination of the de-rotated/re-scaled residuals cube.
    
    If full_output is True, and depending on the type of cube (ADI or ADI+mSDI),
    then several arrays will be returned, such as the residuals, de-rotated
    residuals, principal components,
    """
    if not cube.ndim > 2:
        raise TypeError('Input array is not a 3d or 4d array')

    if check_mem:
        input_bytes = cube.nbytes
        if cube_ref is not None:
            input_bytes += cube_ref.nbytes
        if not check_enough_memory(input_bytes, 1.5, False):
            msgerr = 'Input cubes are larger than available system memory. '
            msgerr += 'Set check_mem=False to override this memory check or '
            msgerr += 'use the incremental PCA (for ADI)'
            raise RuntimeError(msgerr)

    if verbose:
        start_time = time_ini()

    angle_list = check_pa_vector(angle_list)
    #***************************************************************************
    # ADI + mSDI. Shape of cube: (n_channels, n_adi_frames, y, x)
    #***************************************************************************
    if cube.ndim == 4:
        if adimsdi == 'double':
            res_pca = _adimsdi_doublepca(cube, angle_list, scale_list, ncomp,
                                         ncomp2, scaling, mask_center_px, debug,
                                         svd_mode, imlib, interpolation,
                                         collapse, verbose, start_time,
                                         full_output)
            residuals_cube_channels, residuals_cube_channels_, frame = res_pca
        elif adimsdi == 'single':
            res_pca = _adimsdi_singlepca(cube, angle_list, scale_list, ncomp,
                                         scaling, mask_center_px, debug,
                                         svd_mode, imlib, interpolation,
                                         collapse, verbose, start_time,
                                         full_output)
            cube_allfr_residuals, cube_adi_residuals, frame = res_pca
        else:
            raise ValueError('`Adimsdi` mode not recognized')
    # **************************************************************************
    # ADI+RDI
    # **************************************************************************
    elif cube.ndim == 3 and cube_ref is not None:
        res_pca = _adi_rdi_pca(cube, cube_ref, angle_list, ncomp, scaling,
                               mask_center_px, debug, svd_mode, imlib,
                               interpolation, collapse, verbose, full_output,
                               start_time)
        pcs, recon, residuals_cube, residuals_cube_, frame = res_pca
    # **************************************************************************
    # ADI. Shape of cube: (n_adi_frames, y, x)
    # **************************************************************************
    elif cube.ndim == 3 and cube_ref is None:
        res_pca = _adi_pca(cube, angle_list, ncomp, source_xy, delta_rot, fwhm,
                           scaling, mask_center_px, debug, svd_mode, imlib,
                           interpolation, collapse, verbose, start_time, True)

        if source_xy is not None:
            recon_cube, residuals_cube, residuals_cube_, frame = res_pca
        else:
            pcs, recon, residuals_cube, residuals_cube_, frame = res_pca

    else:
        msg = 'Only ADI, ADI+RDI and ADI+mSDI observing techniques are '
        msg += 'supported'
        raise RuntimeError(msg)

    if cube.ndim == 3:
        if full_output:
            if source_xy is not None:
                return recon_cube, residuals_cube, residuals_cube_, frame
            else:
                return pcs, recon, residuals_cube, residuals_cube_, frame
        else:
            return frame
    elif cube.ndim == 4:
        if full_output:
            if adimsdi == 'double':
                return residuals_cube_channels, residuals_cube_channels_, frame
            elif adimsdi == 'single':
                return cube_allfr_residuals, cube_adi_residuals, frame
        else:
            return frame


def _subtr_proj_fullfr(cube, cube_ref, ncomp, scaling, mask_center_px, debug,
                       svd_mode, verbose, full_output, indices=None,
                       frame=None):
    """ PCA projection and model PSF subtraction. Returns the cube of residuals.
    """
    _, y, x = cube.shape
    if indices is not None and frame is not None:
        matrix = prepare_matrix(cube, scaling, mask_center_px, mode='fullfr',
                                verbose=False)
    else:
        matrix = prepare_matrix(cube, scaling, mask_center_px, mode='fullfr',
                                verbose=verbose)

    if cube_ref is not None:
        ref_lib = prepare_matrix(cube_ref, scaling, mask_center_px,
                                 mode='fullfr', verbose=verbose)
    else:
        ref_lib = matrix

    if indices is not None and frame is not None:  # one row (frame) at a time
        ref_lib = ref_lib[indices]
        if ref_lib.shape[0] <= 10:
            msg = 'Too few frames left in the PCA library (<10). '
            msg += 'Try decreasing the parameter delta_rot'
            raise RuntimeError(msg)
        curr_frame = matrix[frame]  # current frame

        V = svd_wrapper(ref_lib, svd_mode, ncomp, False, False)
        transformed = np.dot(curr_frame, V.T)
        reconstructed = np.dot(transformed.T, V)
        residuals = curr_frame - reconstructed
        if full_output:
            return ref_lib.shape[0], residuals, reconstructed
        else:
            return ref_lib.shape[0], residuals
    else:  # the whole matrix
        V = svd_wrapper(ref_lib, svd_mode, ncomp, debug, verbose)
        transformed = np.dot(V, matrix.T)
        reconstructed = np.dot(transformed.T, V)
        residuals = matrix - reconstructed
        residuals_res = reshape_matrix(residuals, y, x)
        if full_output:
            return residuals_res, reconstructed, V
        else:
            return residuals_res


def _adi_pca(cube, angle_list, ncomp, source_xy, delta_rot, fwhm, scaling,
             mask_center_px, debug, svd_mode, imlib, interpolation, collapse,
             verbose, start_time, full_output):
    """ Handles the ADI PCA post-processing.
    """
    n, y, x = cube.shape

    if not n == angle_list.shape[0]:
        msg = "Angle list vector has wrong length. It must equal the "
        msg += "number of frames in the cube"
        raise ValueError(msg)

    if ncomp > n:
        ncomp = min(ncomp, n)
        msg = 'Number of PCs too high (max PCs={}), using {} PCs instead.'
        print(msg.format(n, ncomp))

    if source_xy is None:
        residuals_result = _subtr_proj_fullfr(cube, None, ncomp, scaling,
                                              mask_center_px, debug,
                                              svd_mode, verbose, full_output)
        if verbose:
            timing(start_time)
        if full_output:
            residuals_cube = residuals_result[0]
            reconstructed = residuals_result[1]
            V = residuals_result[2]
            pcs = reshape_matrix(V, y, x)
            recon = reshape_matrix(reconstructed, y, x)
        else:
            residuals_cube = residuals_result
    else:
        if delta_rot is None or fwhm is None:
            msg = 'Delta_rot or fwhm parameters missing. Needed for the'
            msg += 'PA-based rejection of frames from the library'
            raise TypeError(msg)
        nfrslib = []
        residuals_cube = np.zeros_like(cube)
        recon_cube = np.zeros_like(cube)
        yc, xc = frame_center(cube[0], False)
        x1, y1 = source_xy
        ann_center = dist(yc, xc, y1, x1)
        pa_thr = _compute_pa_thresh(ann_center, fwhm, delta_rot)
        mid_range = np.abs(np.amax(angle_list) - np.amin(angle_list)) / 2
        if pa_thr >= mid_range - mid_range * 0.1:
            new_pa_th = float(mid_range - mid_range * 0.1)
            if verbose:
                msg = 'PA threshold {:.2f} is too big, will be set to '
                msg += '{:.2f}'
                print(msg.format(pa_thr, new_pa_th))
            pa_thr = new_pa_th

        for frame in range(n):
            if ann_center > fwhm * 3:  # TODO: 3 optimal value? new par?
                ind = _find_indices(angle_list, frame, pa_thr,
                                    truncate=True)
            else:
                ind = _find_indices(angle_list, frame, pa_thr)

            res_result = _subtr_proj_fullfr(cube, None, ncomp, scaling,
                                            mask_center_px, debug, svd_mode,
                                            verbose, full_output, ind,
                                            frame)
            if full_output:
                nfrslib.append(res_result[0])
                residual_frame = res_result[1]
                recon_frame = res_result[2]
                residuals_cube[frame] = residual_frame.reshape((y, x))
                recon_cube[frame] = recon_frame.reshape((y, x))
            else:
                nfrslib.append(res_result[0])
                residual_frame = res_result[1]
                residuals_cube[frame] = residual_frame.reshape((y, x))

        # number of frames in library printed for each annular quadrant
        if verbose:
            descriptive_stats(nfrslib, verbose=verbose, label='Size LIB: ')

    residuals_cube_ = cube_derotate(residuals_cube, angle_list, imlib=imlib,
                                    interpolation=interpolation)
    frame = cube_collapse(residuals_cube_, mode=collapse)
    if verbose:
        print('Done de-rotating and combining')
        timing(start_time)
    if source_xy is not None:
        return recon_cube, residuals_cube, residuals_cube_, frame
    else:
        return pcs, recon, residuals_cube, residuals_cube_, frame


def _adimsdi_singlepca(cube, angle_list, scale_list, ncomp, scaling,
                       mask_center_px, debug, svd_mode, imlib, interpolation,
                       collapse, verbose, start_time, full_output):
    """ Handles the full-frame ADI+mSDI single PCA post-processing.
    """
    z, n, y_in, x_in = cube.shape

    if ncomp is None:
        raise ValueError('`Ncomp` must be provided (positive integer value)')

    if not angle_list.shape[0] == n:
        msg = "Angle list vector has wrong length. It must equal the number"
        msg += " frames in the cube"
        raise ValueError(msg)

    if scale_list is None:
        raise ValueError('Scaling factors vector must be provided')
    else:
        if np.array(scale_list).ndim > 1:
            raise ValueError('Scaling factors vector is not 1d')
        if not scale_list.shape[0] == cube.shape[0]:
            raise ValueError('Scaling factors vector has wrong length')

    scale_list = check_scal_vector(scale_list)
    big_cube = []

    tit = 'Rescaling the spectral channels to align the speckles'
    bar = pyprind.ProgBar(n, stream=1, title=tit)
    for i in range(n):
        cube_resc, _, _, _, _, _ = scpca(cube[:, i, :, :], scale_list)
        cube_resc = cube_crop_frames(cube_resc, size=y_in, verbose=False)
        big_cube.append(cube_resc)
        bar.update()

    big_cube = np.array(big_cube)
    big_cube = big_cube.reshape(z * n, y_in, x_in)

    if verbose:
        timing(start_time)
        print('{} total frames'.format(n * z))
        print('Performing single PCA')
    res_cube = _subtr_proj_fullfr(big_cube, None, ncomp, scaling,
                                  mask_center_px, debug, svd_mode, False, False)

    if verbose:
        timing(start_time)

    resadi_cube = np.zeros((n, y_in, x_in))
    tit = 'Descaling the spectral channels'
    bar = pyprind.ProgBar(n, stream=1, title=tit)
    for i in range(n):
        frame_i = scpca(res_cube[i * z:(i+1) * z, :, :], scale_list,
                        full_output=full_output, inverse=True, y_in=y_in,
                        x_in=x_in)
        resadi_cube[i] = frame_i
        bar.update()

    if verbose:
        print('De-rotating and combining residuals')
        timing(start_time)
    der_res = cube_derotate(resadi_cube, angle_list, imlib=imlib,
                            interpolation=interpolation)
    frame = cube_collapse(der_res, mode=collapse)

    cube_allfr_residuals = res_cube
    cube_adi_residuals = resadi_cube
    return cube_allfr_residuals, cube_adi_residuals, frame


def _adimsdi_doublepca(cube, angle_list, scale_list, ncomp, ncomp2, scaling,
                       mask_center_px, debug, svd_mode, imlib, interpolation,
                       collapse, verbose, start_time, full_output):
    """  Handles the full-frame ADI+mSDI double PCA post-processing.
    """
    z, n, y_in, x_in = cube.shape

    if not angle_list.shape[0] == n:
        msg = "Angle list vector has wrong length. It must equal the number"
        msg += " frames in the cube"
        raise ValueError(msg)

    if scale_list is None:
        raise ValueError('Scaling factors vector must be provided')
    else:
        if np.array(scale_list).ndim > 1:
            raise ValueError('Scaling factors vector is not 1d')
        if not scale_list.shape[0] == cube.shape[0]:
            raise ValueError('Scaling factors vector has wrong length')

    scale_list = check_scal_vector(scale_list)

    if verbose:
        print('{} spectral channels in IFS cube'.format(z))
    residuals_cube_channels = np.zeros((n, y_in, x_in))

    if ncomp is None:
        tit = 'Combining multi-spectral frames (skipping PCA)'
        bar = pyprind.ProgBar(n, stream=1, title=tit)
        for i in range(n):
            frame_i = cube_collapse(cube[:, i, :, :], mode=collapse)
            residuals_cube_channels[i] = frame_i
            bar.update()
    else:
        if ncomp > z:
            ncomp = min(ncomp, z)
            msg = 'Number of PCs too high (max PCs={}), using {} PCs instead'
            print(msg.format(z, ncomp))

        tit = 'First PCA stage exploiting spectral variability'
        bar = pyprind.ProgBar(n, stream=1, title=tit)
        for i in range(n):
            cube_resc, _, _, _, _, _ = scpca(cube[:, i, :, :], scale_list)
            residuals_result = _subtr_proj_fullfr(cube_resc, None, ncomp,
                                                  scaling, mask_center_px,
                                                  debug, svd_mode, False,
                                                  full_output)
            if full_output:
                residuals_cube = residuals_result[0]
            else:
                residuals_cube = residuals_result
            frame_i = scpca(residuals_cube, scale_list, full_output=full_output,
                            inverse=True, y_in=y_in, x_in=x_in)
            residuals_cube_channels[i] = frame_i
            bar.update()

    if verbose:
        timing(start_time)

    # de-rotation of the PCA processed channels, ADI fashion
    if ncomp2 is None:
        if verbose:
            print('{} ADI frames'.format(n))
            print('De-rotating and combining frames (skipping PCA)')
        residuals_cube_channels_ = cube_derotate(residuals_cube_channels,
                                                 angle_list, imlib=imlib,
                                                 interpolation=interpolation)
        frame = cube_collapse(residuals_cube_channels_, mode=collapse)
        if verbose:
            timing(start_time)
    else:
        if ncomp2 > n:
            ncomp2 = min(ncomp2, n)
            msg = 'Number of PCs too high (max PCs={}), using {} PCs '
            msg += 'instead'
            print(msg.format(n, ncomp2))
        if verbose:
            print('{} ADI frames'.format(n))
            msg = 'Second PCA stage exploiting rotational variability'
            print(msg)
        res_ifs_adi = _subtr_proj_fullfr(residuals_cube_channels, None,
                                         ncomp2, scaling, mask_center_px,
                                         debug, svd_mode, False,
                                         full_output)
        if verbose:
            print('De-rotating and combining residuals')
        der_res = cube_derotate(res_ifs_adi, angle_list, imlib=imlib,
                                interpolation=interpolation)
        residuals_cube_channels_ = der_res
        frame = cube_collapse(residuals_cube_channels_, mode=collapse)
        if verbose:
            timing(start_time)
    return residuals_cube_channels, residuals_cube_channels_, frame


def _adi_rdi_pca(cube, cube_ref, angle_list, ncomp, scaling, mask_center_px,
                 debug, svd_mode, imlib, interpolation, collapse, verbose,
                 full_output, start_time):
    """ Handles the ADI+RDI post-processing.
    """
    n, y, x = cube.shape
    if not cube_ref.ndim == 3:
        msg = 'Input reference array is not a cube or 3d array'
        raise ValueError(msg)
    if not cube_ref.shape[1] == y:
        msg = 'Reference and target frames have different shape'
        raise TypeError(msg)

    if ncomp > n:
        ncomp = min(ncomp, n)
        msg = 'Number of PCs too high (max PCs={}), using {} PCs instead.'
        print(msg.format(n, ncomp))
    residuals_result = _subtr_proj_fullfr(cube, cube_ref, ncomp, scaling,
                                          mask_center_px, debug, svd_mode,
                                          verbose, full_output)
    if full_output:
        residuals_cube = residuals_result[0]
        reconstructed = residuals_result[1]
        V = residuals_result[2]
        pcs = reshape_matrix(V, y, x)
        recon = reshape_matrix(reconstructed, y, x)
    else:
        residuals_cube = residuals_result
    residuals_cube_ = cube_derotate(residuals_cube, angle_list, imlib=imlib,
                                    interpolation=interpolation)
    frame = cube_collapse(residuals_cube_, mode=collapse)

    if verbose:
        print('Done de-rotating and combining')
        timing(start_time)
    return pcs, recon, residuals_cube, residuals_cube_, frame


def pca_optimize_snr(cube, angle_list, source_xy, fwhm, cube_ref=None,
                     mode='fullfr', annulus_width=20, range_pcs=None,
                     svd_mode='lapack', scaling=None, mask_center_px=None, 
                     fmerit='px', min_snr=0, imlib='opencv',
                     interpolation='lanczos4', collapse='median', verbose=True,
                     full_output=False, debug=False, plot=True, save_plot=None,
                     plot_title=None):
    """ Optimizes the number of principal components by doing a simple grid 
    search measuring the SNR for a given position in the frame (ADI, RDI). 
    The metric used could be the given pixel's SNR, the maximum SNR in a FWHM 
    circular aperture centered on the given coordinates or the mean SNR in the 
    same circular aperture. They yield slightly different results.
    
    Parameters
    ----------
    cube : array_like, 3d
        Input cube.
    angle_list : array_like, 1d
        Corresponding parallactic angle for each frame.
    source_xy : tuple of floats
        X and Y coordinates of the pixel where the source is located and whose
        SNR is going to be maximized.
    fwhm : float 
        Size of the PSF's FWHM in pixels.
    cube_ref : array_like, 3d, optional
        Reference library cube. For Reference Star Differential Imaging. 
    mode : {'fullfr', 'annular'}, optional
        Mode for PCA processing (full-frame or just in an annulus). There is a
        catch: the optimal number of PCs in full-frame may not coincide with the
        one in annular mode. This is due to the fact that the annulus matrix is
        smaller (less noisy, probably not containing the central star) and also
        its intrinsic rank (smaller that in the full frame case).
    annulus_width : float, optional
        Width in pixels of the annulus in the case of the "annular" mode. 
    range_pcs : tuple, optional
        The interval of PCs to be tried. If None then the algorithm will find
        a clever way to sample from 1 to 200 PCs. If a range is entered (as 
        (PC_INI, PC_MAX)) a sequential grid will be evaluated between PC_INI
        and PC_MAX with step of 1. If a range is entered (as 
        (PC_INI, PC_MAX, STEP)) a grid will be evaluated between PC_INI and 
        PC_MAX with the given STEP.          
    svd_mode : {'lapack', 'arpack', 'eigen', 'randsvd', 'cupy', 'eigencupy', 'randcupy'}, str
        Switch for the SVD method/library to be used. ``lapack`` uses the LAPACK 
        linear algebra library through Numpy and it is the most conventional way 
        of computing the SVD (deterministic result computed on CPU). ``arpack`` 
        uses the ARPACK Fortran libraries accessible through Scipy (computation
        on CPU). ``eigen`` computes the singular vectors through the 
        eigendecomposition of the covariance M.M' (computation on CPU).
        ``randsvd`` uses the randomized_svd algorithm implemented in Sklearn 
        (computation on CPU). ``cupy`` uses the Cupy library for GPU computation
        of the SVD as in the LAPACK version. ``eigencupy`` offers the same 
        method as with the ``eigen`` option but on GPU (through Cupy). 
        ``randcupy`` is an adaptation of the randomized_svd algorith, where all 
        the computations are done on a GPU. 
    scaling : {None, 'temp-mean', 'spat-mean', 'temp-standard', 'spat-standard'}
        With None, no scaling is performed on the input data before SVD. With 
        "temp-mean" then temporal px-wise mean subtraction is done, with 
        "spat-mean" then the spatial mean is subtracted, with "temp-standard" 
        temporal mean centering plus scaling to unit variance is done and with
        "spat-standard" spatial mean centering plus scaling to unit variance is
        performed.  
    mask_center_px : None or int, optional
        If None, no masking is done. If an integer > 1 then this value is the
        radius of the circular mask.  
    fmerit : {'px', 'max', 'mean'}
        The function of merit to be maximized. 'px' is *source_xy* pixel's SNR, 
        'max' the maximum SNR in a FWHM circular aperture centered on 
        *source_xy* and 'mean' is the mean SNR in the same circular aperture.  
    min_snr : float
        Value for the minimum acceptable SNR. Setting this value higher will 
        reduce the steps.
    imlib : str, optional
        See the documentation of the ``vip_hci.preproc.frame_rotate`` function.
    interpolation : str, optional
        See the documentation of the ``vip_hci.preproc.frame_rotate`` function.
    collapse : {'median', 'mean', 'sum', 'trimmean'}, str optional
        Sets the way of collapsing the frames for producing a final image.
    verbose : {True, False}, bool optional
        If True prints intermediate info and timing.
    full_output : {False, True} bool optional
        If True it returns the optimal number of PCs, the final PCA frame for 
        the optimal PCs and a cube with all the final frames for each number 
        of PC that was tried.
    debug : {False, True}, bool optional
        Whether to print debug information or not.
    plot : {True, False}, optional
        Whether to plot the SNR and flux as functions of PCs and final PCA 
        frame or not.
    save_plot: string
        If provided, the pc optimization plot will be saved to that path.
    plot_title: string
        If provided, the plot is titled

    Returns
    -------
    opt_npc : int
        Optimal number of PCs for given source.
    If full_output is True, the final processed frame, and a cube with all the
    PCA frames are returned along with the optimal number of PCs.
    """    
    def truncate_svd_get_finframe(matrix, angle_list, ncomp, V):
        """ Projection, subtraction, derotation plus combination in one frame.
        Only for full-frame"""
        transformed = np.dot(V[:ncomp], matrix.T)
        reconstructed = np.dot(transformed.T, V[:ncomp])
        residuals = matrix - reconstructed
        frsize = int(np.sqrt(matrix.shape[1]))          # only for square frames
        residuals_res = reshape_matrix(residuals, frsize, frsize)
        residuals_res_der = cube_derotate(residuals_res, angle_list,
                                          imlib=imlib,
                                          interpolation=interpolation)
        frame = cube_collapse(residuals_res_der, mode=collapse)
        return frame

    def truncate_svd_get_finframe_ann(matrix, indices, angle_list, ncomp, V):
        """ Projection, subtraction, derotation plus combination in one frame.
        Only for annular mode"""
        transformed = np.dot(V[:ncomp], matrix.T)
        reconstructed = np.dot(transformed.T, V[:ncomp])
        residuals_ann = matrix - reconstructed
        residuals_res = np.zeros_like(cube)
        residuals_res[:,indices[0],indices[1]] = residuals_ann
        residuals_res_der = cube_derotate(residuals_res, angle_list,
                                          imlib=imlib,
                                          interpolation=interpolation)
        frame = cube_collapse(residuals_res_der, mode=collapse)
        return frame

    def get_snr(matrix, angle_list, y, x, mode, V, fwhm, ncomp, fmerit,
                full_output):
        if mode == 'fullfr':
            frame = truncate_svd_get_finframe(matrix, angle_list, ncomp, V)
        elif mode == 'annular':
            frame = truncate_svd_get_finframe_ann(matrix, annind, angle_list,
                                                  ncomp, V)
        else:
            raise RuntimeError('Wrong mode. Choose either full or annular')
        
        if fmerit == 'max':
            yy, xx = draw.circle(y, x, fwhm/2.)
            res = [phot.snr_ss(frame, (x_,y_), fwhm, plot=False, verbose=False, 
                               full_output=True) for y_, x_ in zip(yy, xx)]
            snr_pixels = np.array(res)[:,-1]
            fluxes = np.array(res)[:,2]
            argm = np.argmax(snr_pixels)
            if full_output:
                # integrated fluxes for the max snr
                return np.max(snr_pixels), fluxes[argm], frame
            else:
                return np.max(snr_pixels), fluxes[argm]
        elif fmerit == 'px':
            res = phot.snr_ss(frame, (x,y), fwhm, plot=False, verbose=False,
                              full_output=True)
            snrpx = res[-1]
            fluxpx = np.array(res)[2]
            if full_output:
                # integrated fluxes for the given px
                return snrpx, fluxpx, frame
            else:
                return snrpx, fluxpx
        elif fmerit == 'mean':
            yy, xx = draw.circle(y, x, fwhm/2.)
            res = [phot.snr_ss(frame, (x_,y_), fwhm, plot=False, verbose=False, 
                               full_output=True) for y_, x_ in zip(yy, xx)]  
            snr_pixels = np.array(res)[:,-1]
            fluxes = np.array(res)[:,2]
            if full_output:
                # mean of the integrated fluxes (shifting the aperture)
                return np.mean(snr_pixels), np.mean(fluxes), frame
            else:                         
                return np.mean(snr_pixels), np.mean(fluxes)
    
    def grid(matrix, angle_list, y, x, mode, V, fwhm, fmerit, step, inti, intf, 
             debug, full_output, truncate=True):
        nsteps = 0
        snrlist = []
        pclist = []
        fluxlist = []
        if full_output:
            frlist = []
        counter = 0
        if debug:  
            print('Step current grid:', step)
            print('PCs | SNR')
        for pc in range(inti, intf+1, step):
            if full_output:
                snr, flux, frame = get_snr(matrix, angle_list, y, x, mode, V,
                                           fwhm, pc, fmerit, full_output)
            else:
                snr, flux = get_snr(matrix, angle_list, y, x, mode, V, fwhm, pc,
                                    fmerit, full_output)
            if np.isnan(snr):
                snr = 0
            if nsteps > 1 and snr < snrlist[-1]:
                counter += 1
            snrlist.append(snr)
            pclist.append(pc)
            fluxlist.append(flux)
            if full_output:
                frlist.append(frame)
            nsteps += 1
            if truncate and nsteps > 2 and snr < min_snr:
                if debug:
                    print('SNR too small')
                break
            if debug:
                print('{} {:.3f}'.format(pc, snr))
            if truncate and counter == 5:
                break
        argm = np.argmax(snrlist)
        
        if len(pclist) == 2:
            pclist.append(pclist[-1]+1)
    
        if debug:
            print('Finished current stage')
            try:
                pclist[argm+1]
                print('Interval for next grid: ', pclist[argm-1], 'to',
                      pclist[argm+1])
            except:
                print('The optimal SNR seems to be outside of the given '
                      'PC range')
            print()
        
        if argm == 0:
            argm = 1
        if full_output:
            return argm, pclist, snrlist, fluxlist, frlist
        else: 
            return argm, pclist, snrlist, fluxlist
    
    #---------------------------------------------------------------------------
    if cube.ndim != 3:
        raise TypeError('Input array is not a cube or 3d array')
    
    if verbose:
        start_time = time_ini()
    n = cube.shape[0]
    x, y = source_xy 
    
    if range_pcs is not None:
        if len(range_pcs) == 2:
            pcmin, pcmax = range_pcs
            pcmax = min(pcmax, n)
            step = 1
        elif len(range_pcs) == 3:
            pcmin, pcmax, step = range_pcs
            pcmax = min(pcmax, n)
        else:
            msg = 'Range_pcs tuple must be entered as (PC_INI, PC_MAX, STEP) '
            msg += 'or (PC_INI, PC_MAX)'
            raise TypeError(msg)
    else:
        pcmin = 1
        pcmax = 200
        pcmax = min(pcmax, n)
    
    # Getting `pcmax` principal components a single time
    if mode == 'fullfr':
        matrix = prepare_matrix(cube, scaling, mask_center_px, verbose=False)
        if cube_ref is not None:
            ref_lib = prepare_matrix(cube_ref, scaling, mask_center_px,
                                     verbose=False)
        else:
            ref_lib = matrix

    elif mode == 'annular':
        y_cent, x_cent = frame_center(cube[0])
        ann_radius = dist(y_cent, x_cent, y, x)
        matrix, annind = prepare_matrix(cube, scaling, None, mode='annular',
                                        annulus_radius=ann_radius,
                                        annulus_width=annulus_width,
                                        verbose=False)
        if cube_ref is not None:
            ref_lib, _ = prepare_matrix(cube_ref, scaling, mask_center_px,
                                     mode='annular', annulus_radius=ann_radius,
                                     annulus_width=annulus_width, verbose=False)
        else:
            ref_lib = matrix

    else:
        raise RuntimeError('Wrong mode. Choose either fullfr or annular')

    V = svd_wrapper(ref_lib, svd_mode, pcmax, False, verbose)

    # sequential grid
    if range_pcs is not None:
        grid1 = grid(matrix, angle_list, y, x, mode, V, fwhm, fmerit, step, 
                     pcmin, pcmax, debug, full_output, False)
        if full_output:
            argm, pclist, snrlist, fluxlist, frlist = grid1
        else:
            argm, pclist, snrlist, fluxlist = grid1
        
        opt_npc = pclist[argm]    
        if verbose:
            print('Number of steps', len(pclist))
            msg = 'Optimal number of PCs = {}, for SNR={:.3f}'
            print(msg.format(opt_npc, snrlist[argm]))
            print()
            timing(start_time)
        
        if full_output: 
            cubeout = np.array((frlist))

        # Plot of SNR as function of PCs  
        if plot:    
            plt.figure(figsize=(8, 4))
            ax1 = plt.subplot(211)     
            ax1.plot(pclist, snrlist, '-', alpha=0.5)
            ax1.plot(pclist, snrlist, 'o', alpha=0.5, color='blue')
            ax1.set_xlim(np.array(pclist).min(), np.array(pclist).max())
            ax1.set_ylim(0, np.array(snrlist).max()+1)
            ax1.set_ylabel('SNR')
            ax1.minorticks_on()
            ax1.grid('on', 'major', linestyle='solid', alpha=0.4)
            
            ax2 = plt.subplot(212)
            ax2.plot(pclist, fluxlist, '-', alpha=0.5, color='green')
            ax2.plot(pclist, fluxlist, 'o', alpha=0.5, color='green')
            ax2.set_xlim(np.array(pclist).min(), np.array(pclist).max())
            ax2.set_ylim(0, np.array(fluxlist).max()+1)
            ax2.set_xlabel('Principal components')
            ax2.set_ylabel('Flux in FWHM ap. [ADUs]')
            ax2.minorticks_on()
            ax2.grid('on', 'major', linestyle='solid', alpha=0.4)
            print()
            
    # automatic "clever" grid
    else:
        grid1 = grid(matrix, angle_list, y, x, mode, V, fwhm, fmerit, 
                     max(int(pcmax*0.1),1), pcmin, pcmax, debug, full_output)
        if full_output:
            argm, pclist, snrlist, fluxlist, frlist1 = grid1
        else:
            argm, pclist, snrlist, fluxlist = grid1
        
        grid2 = grid(matrix, angle_list, y, x, mode, V, fwhm, fmerit, 
                     max(int(pcmax*0.05),1), pclist[argm-1], pclist[argm+1],
                     debug, full_output)
        if full_output:
            argm2, pclist2, snrlist2, fluxlist2, frlist2 = grid2
        else:
            argm2, pclist2, snrlist2, fluxlist2 = grid2
        
        grid3 = grid(matrix, angle_list, y, x, mode, V, fwhm, fmerit, 1, 
                     pclist2[argm2-1], pclist2[argm2+1], debug, full_output, 
                     False)
        if full_output:
            _, pclist3, snrlist3, fluxlist3, frlist3 = grid3
        else:
            _, pclist3, snrlist3, fluxlist3 = grid3
        
        argm = np.argmax(snrlist3)
        opt_npc = pclist3[argm]    
        dfr = pd.DataFrame(np.array((pclist+pclist2+pclist3, 
                                     snrlist+snrlist2+snrlist3,
                                     fluxlist+fluxlist2+fluxlist3)).T)  
        dfrs = dfr.sort_values(0)
        dfrsrd = dfrs.drop_duplicates()
        ind = np.array(dfrsrd.index)    
        
        if verbose:
            print('Number of evaluated steps', ind.shape[0])
            msg = 'Optimal number of PCs = {}, for SNR={:.3f}'
            print(msg.format(opt_npc, snrlist3[argm]), '\n')
            timing(start_time)
        
        if full_output: 
            cubefrs = np.array((frlist1+frlist2+frlist3))
            cubeout = cubefrs[ind]
    
        # Plot of SNR as function of PCs  
        if plot:   
            alpha = 0.4
            lw = 2
            plt.figure(figsize=(6,4))   
            ax1 = plt.subplot(211)  
            ax1.plot(np.array(dfrsrd.loc[:,0]), np.array(dfrsrd.loc[:,1]), '-', 
                     alpha=alpha, color='blue', lw=lw)
            ax1.plot(np.array(dfrsrd.loc[:,0]), np.array(dfrsrd.loc[:,1]), 'o',  
                     alpha=alpha/2, color='blue')
            ax1.set_xlim(np.array(dfrsrd.loc[:,0]).min(),
                         np.array(dfrsrd.loc[:,0]).max())
            ax1.set_ylim(0, np.array(dfrsrd.loc[:,1]).max()+1)
            ax1.set_ylabel('S/N')
            ax1.minorticks_on()
            ax1.grid('on', 'major', linestyle='solid', alpha=0.2)
            if plot_title is not None:
                ax1.set_title('Optimal pc: {} for {}'.format(opt_npc,
                                                             plot_title))
            
            ax2 = plt.subplot(212)
            ax2.plot(np.array(dfrsrd.loc[:,0]), np.array(dfrsrd.loc[:,2]), '-', 
                     alpha=alpha, color='green', lw=lw)
            ax2.plot(np.array(dfrsrd.loc[:,0]), np.array(dfrsrd.loc[:,2]), 'o', 
                     alpha=alpha/2, color='green')
            ax2.set_xlim(np.array(pclist).min(), np.array(pclist).max())
            #ax2.set_ylim(0, np.array(fluxlist).max()+1)
            ax2.set_xlabel('Principal components')
            ax2.set_ylabel('Flux in FWHM aperture')
            ax2.minorticks_on()
            ax2.set_yscale('log')
            ax2.grid('on', 'major', linestyle='solid', alpha=0.2)
            print()
    
    # Optionally, save the contrast curve
    if save_plot is not None:
        plt.savefig(save_plot, dpi=100, bbox_inches='tight')

    if mode == 'fullfr':
        finalfr = pca(cube, angle_list, cube_ref=cube_ref, ncomp=opt_npc,
                      svd_mode=svd_mode, mask_center_px=mask_center_px,
                      scaling=scaling, imlib=imlib, interpolation=interpolation,
                      collapse=collapse, verbose=False)
    elif mode == 'annular':
        finalfr = pca_annulus(cube, angle_list, ncomp=opt_npc,
                              annulus_width=annulus_width, r_guess=ann_radius,
                              cube_ref=cube_ref, svd_mode=svd_mode,
                              scaling=scaling, collapse=collapse, imlib=imlib,
                              interpolation=interpolation)

    _ = phot.frame_quick_report(finalfr, fwhm, (x,y), verbose=verbose)
    
    if full_output:
        return opt_npc, finalfr, cubeout
    else:
        return opt_npc


def pca_incremental(cubepath, angle_list=None, n=0, batch_size=None, 
                    batch_ratio=0.1, ncomp=10, imlib='opencv',
                    interpolation='lanczos4', collapse='median',
                    verbose=True, full_output=False):
    """ Computes the full-frame PCA-ADI algorithm in batches, for processing 
    fits files larger than the available system memory. It uses the incremental 
    PCA algorithm from scikit-learn. 
    
    Parameters
    ----------
    cubepath : str
        String with the path to the fits file to be opened in memmap mode.
    angle_list : array_like, 1d
        Corresponding parallactic angle for each frame. If None the parallactic
        angles are obtained from the same fits file (extension). 
    n : int optional
        The index of the HDULIST contaning the data/cube.
    batch_size : int optional
        The number of frames in each batch. If None the size of the batch is 
        computed wrt the available memory in the system.
    batch_ratio : float
        If batch_size is None, batch_ratio indicates the % of the available 
        memory that should be used by every batch.
    ncomp : int, optional
        How many PCs are used as a lower-dimensional subspace to project the
        target frames.
    imlib : str, optional
        See the documentation of the ``vip_hci.preproc.frame_rotate`` function.
    interpolation : str, optional
        See the documentation of the ``vip_hci.preproc.frame_rotate`` function.
    collapse : {'median', 'mean', 'sum', 'trimmean'}, str optional
        Sets the way of collapsing the frames for producing a final image.
    verbose : {True, False}, bool optional
        If True prints intermediate info and timing. 
    full_output: boolean, optional
        Whether to return the final median combined image only or with other 
        intermediate arrays.  
        
    Returns
    -------
    If full_output is True the algorithm returns the incremental PCA model of
    scikit-learn, the PCs reshaped into images, the median of the derotated 
    residuals for each batch, and the final frame. If full_output is False then
    the final frame is returned.
    
    """
    if verbose:  start = time_ini()
    if not isinstance(cubepath, str):
        raise TypeError('Cubepath must be a string with the full path of your '
                        'fits file')
      
    fitsfilename = cubepath
    hdulist = fits.open(fitsfilename, memmap=True)
    if not hdulist[n].data.ndim>2:
        raise TypeError('Input array is not a 3d or 4d array')
    
    n_frames = hdulist[n].data.shape[0]
    y = hdulist[n].data.shape[1]
    x = hdulist[n].data.shape[2]
    if angle_list is None:
        try:  
            angle_list = hdulist[n+1].data
        except:  
            raise RuntimeError('Parallactic angles were not provided')
    if not n_frames==angle_list.shape[0]:
        raise TypeError('Angle list vector has wrong length. It must equal the '
                        'number of frames in the cube.')
    
    ipca = IncrementalPCA(n_components=ncomp)
    
    if batch_size is None:
        aval_mem = get_available_memory(verbose)
        total_size = hdulist[n].data.nbytes
        batch_size = int(n_frames/(total_size/(batch_ratio*aval_mem)))
    
    if verbose:
        msg1 = "Cube with {} frames ({:.3f} GB)"
        print(msg1.format(n_frames, hdulist[n].data.nbytes/1e9))
        msg2 = "Batch size set to {} frames ({:.3f} GB)\n"
        print(msg2.format(batch_size, hdulist[n].data[:batch_size].nbytes/1e9))
                
    res = n_frames % batch_size
    for i in range(0, n_frames//batch_size):
        intini = i*batch_size
        intfin = (i+1)*batch_size
        batch = hdulist[n].data[intini:intfin]
        msg = 'Processing batch [{},{}] with shape {}'
        if verbose:
            print(msg.format(intini, intfin, batch.shape))
            print('Batch size in memory = {:.3f} MB'.format(batch.nbytes/1e6))
        matrix = prepare_matrix(batch, verbose=False)
        ipca.partial_fit(matrix)
    if res > 0:
        batch = hdulist[n].data[intfin:]
        msg = 'Processing batch [{},{}] with shape {}'
        if verbose:
            print(msg.format(intfin, n_frames, batch.shape))
            print('Batch size in memory = {:.3f} MB'.format(batch.nbytes/1e6))
        matrix = prepare_matrix(batch, verbose=False)
        ipca.partial_fit(matrix)
    
    if verbose:
        timing(start)
    
    V = ipca.components_
    mean = ipca.mean_.reshape(batch.shape[1], batch.shape[2])
    
    if verbose:
        print('\nReconstructing and obtaining residuals')
    medians = []
    for i in range(0, n_frames//batch_size):
        intini = i*batch_size
        intfin = (i+1)*batch_size
        batch = hdulist[n].data[intini:intfin]
        batch = batch - mean 
        matrix = prepare_matrix(batch, verbose=False)
        reconst = np.dot(np.dot(matrix, V.T), V)
        resid = matrix - reconst
        resid_der = cube_derotate(resid.reshape(batch.shape[0], 
                                                batch.shape[1],
                                                batch.shape[2]), 
                                  angle_list[intini:intfin], imlib=imlib,
                                  interpolation=interpolation)
        medians.append(cube_collapse(resid_der, mode=collapse))
    if res > 0:
        batch = hdulist[n].data[intfin:]
        batch = batch - mean
        matrix = prepare_matrix(batch, verbose=False)
        reconst = np.dot(np.dot(matrix, V.T), V)
        resid = matrix - reconst
        resid_der = cube_derotate(resid.reshape(batch.shape[0], 
                                                batch.shape[1],
                                                batch.shape[2]), 
                                  angle_list[intfin:], imlib=imlib,
                                  interpolation=interpolation)
        medians.append(cube_collapse(resid_der, mode=collapse))
    del matrix
    del batch

    medians = np.array(medians)
    frame = np.median(medians, axis=0)
    
    if verbose:
        timing(start)

    if full_output:
        pcs = reshape_matrix(V, y, x)
        return ipca, pcs, medians, frame
    else:
        return frame
    
    



#! /usr/bin/env python

"""
Module with various functions.
"""

from __future__ import division, print_function

__author__ = 'Carlos Alberto Gomez Gonzalez'
__all__ = ['dist',
           'frame_center',
           'get_square',
           'get_circle',
           'get_ellipse',
           'get_annulus_segments',
           'get_annulus',
           'get_annulus_cube',
           'get_ell_annulus',
           'mask_circle',
           'create_ringed_spider_mask',
           'matrix_scaling',
           'prepare_matrix',
           'reshape_matrix']

import numpy as np
from skimage.draw import polygon
from skimage.draw import circle
from sklearn.preprocessing import scale


def mask_circle(array, radius, fillwith=0, mode='in'):
    """ Masks the pixels inside/outside (depending on ``mode``) of a centered
    circle from a frame or cube, replacing the values with the value of
    ``fillwith``.

    Parameters
    ----------
    array : array_like
        Input frame or cube.
    radius : int
        Radius of the circular aperture.
    fillwith : int, float or np.nan, optional
        Value to put instead of the masked out pixels.
    mode : {'in', 'out'}, optional
        When set to 'in' then the pixels inside the radius are set to
        ``fillwith``. When set to 'out' the pixels outside the circular mask are
        set to ``fillwith``.

    Returns
    -------
    array_masked : array_like
        Masked frame or cube.
    """
    if not isinstance(fillwith, (int, float)):
        raise ValueError('`Fillwith` must be integer, float or np.nan')

    if array.ndim == 2:
        cy, cx = frame_center(array)
    elif array.ndim == 3:
        cy, cx = frame_center(array[0])
    ind = circle(cy, cx, radius)

    if mode == 'in':
        array_masked = array.copy()
    elif mode == 'out':
        array_masked = np.ones_like(array) * np.nan

    if mode == 'in':
        if array.ndim == 2:
            array_masked[ind] = fillwith
        elif array.ndim == 3:
            for i in range(array.shape[0]):
                array_masked[i][ind] = fillwith

    elif mode == 'out':
        if array.ndim == 2:
            array_masked[ind] = array[ind]
        elif array.ndim == 3:
            for i in range(array.shape[0]):
                array_masked[i][ind] = array[i][ind]

    return array_masked


def create_ringed_spider_mask(im_shape, ann_out, ann_in=0, sp_width=10,
                              sp_angle=0):
    """
    Mask out information is outside the annulus and inside the spiders (zeros).

    Parameters
    ----------
    im_shape : tuple of int
        Tuple of length two with 2d array shape (Y,X).
    ann_out : int
        Outer radius of the annulus.
    ann_in : int
        Inner radius of the annulus.
    sp_width : int
        Width of the spider arms (3 branches).
    sp_angle : int
        angle of the first spider arm (on the positive horizontal axis) in
        counter-clockwise sense.

    Returns
    -------
    mask : array_like
        2d array of zeros and ones.

    """
    mask = np.zeros(im_shape)

    s = im_shape[0]
    r = s/2
    theta = np.arctan2(sp_width/2, r)

    t0 = np.array([theta, np.pi-theta, np.pi+theta, np.pi*2 - theta])
    t1 = t0 + sp_angle/180 * np.pi
    t2 = t1 + np.pi/3
    t3 = t2 + np.pi/3

    x1 = r * np.cos(t1) + s/2
    y1 = r * np.sin(t1) + s/2
    x2 = r * np.cos(t2) + s/2
    y2 = r * np.sin(t2) + s/2
    x3 = r * np.cos(t3) + s/2
    y3 = r * np.sin(t3) + s/2

    rr1, cc1 = polygon(y1, x1)
    rr2, cc2 = polygon(y2, x2)
    rr3, cc3 = polygon(y3, x3)

    cy, cx = frame_center(mask)
    rr0, cc0 = circle(cy, cx, min(ann_out, cy))
    rr4, cc4 = circle(cy, cx, ann_in)

    mask[rr0, cc0] = 1
    mask[rr1, cc1] = 0
    mask[rr2, cc2] = 0
    mask[rr3, cc3] = 0
    mask[rr4, cc4] = 0
    return mask


def dist(yc,xc,y1,x1):
    """ Returns the Euclidean distance between two points.
    """
    return np.sqrt((yc-y1)**2 + (xc-x1)**2)


def frame_center(array, verbose=False):
    """ Returns the coordinates y,x of an image center.
    """   
    cy = array.shape[0]/2 - 0.5
    cx = array.shape[1]/2 - 0.5

    if verbose:
        print('Center px coordinates at x,y = ({}, {})'.format(cy, cx))
    return cy, cx

    
def get_square(array, size, y, x, position=False, force=False):
    """ Returns an square subframe from a 2d array or image.
    
    Parameters
    ----------
    array : array_like
        Input frame.
    size : int
        Size of the subframe.
    y : int
        Y coordinate of the center of the subframe (obtained with the function
        ``frame_center``).
    x : int
        X coordinate of the center of the subframe (obtained with the function
        ``frame_center``).
    position : bool, optional
        If set to True return also the coordinates of the bottom-left vertex.
    force : bool, optional
        Size and the size of the 2d array must be both even or odd. With
        ``force`` set to True this condition can be avoided.
        
    Returns
    -------
    array_view : array_like
        Sub array.
        
    """
    if array.ndim != 2:
        raise TypeError('Input array is not a frame or 2d array.')

    ary, arx = array.shape

    if not force:
        if ary % 2 == 0:    # assuming square frames
            if size % 2 != 0:
                size += 1
                print('`Size` is odd (while frame size is even)')
                print('Setting `size` to {} pixels'.format(size))
        else:
            if size % 2 == 0:
                size += 1
                print('`Size` is even (while frame size is odd)')
                print('Setting `size` to {} pixels'.format(size))

    # wing is added to the sides of the subframe center
    if size % 2 != 0:
        wing = size//2
    else:
        wing = size/2 - 0.5

    y0 = int(y-wing)
    y1 = int(y+wing+1)  # +1 cause endpoint is excluded when slicing
    x0 = int(x-wing)
    x1 = int(x+wing+1)
    if (y0 or x0) < 0 or (y1 or x1) > ary:   # assuming square frames
        msg = 'square cannot be obtained given the size and y,x combination'
        raise RuntimeError(msg)

    array_view = array[y0: y1, x0: x1].copy()
    
    if position:
        return array_view, y0, x0
    else:
        return array_view


def get_circle(array, radius, output_values=False, cy=None, cx=None):
    """Returns a centered circular region from a 2d ndarray. All the rest
    pixels are set to zeros.

    Parameters
    ----------
    array : array_like
        Input 2d array or image.
    radius : int
        The radius of the circular region.
    output_values : bool, optional
        Sets the type of output.
    cy, cx : int, optional
        Coordinates of the circle center.

    Returns
    -------
    values : array_like
        1d array with the values of the pixels in the circular region.
    array_masked : array_like
        Input array with the circular mask applied.
    """
    if array.ndim != 2:
        raise TypeError('Input array is not a frame or 2d array.')
    sy, sx = array.shape
    if cy is None or cx is None:
        cy, cx = frame_center(array, verbose=False)

    # ogrid is a multidim mesh creator (faster than mgrid)
    yy, xx = np.ogrid[:sy,:sx]
    circle = (yy - cy) ** 2 + (xx - cx) ** 2  # eq of circle. sq dist to center
    circle_mask = circle < radius ** 2  # mask of 1's and 0's
    if output_values:
        values = array[circle_mask]
        return values
    else:
        array_masked = array * circle_mask
        return array_masked


def get_ellipse(array, a, b, PA, output_values=False, cy=None, cx=None,
                output_indices=False):
    """ Returns a centered elliptical region from a 2d ndarray. All the rest 
    pixels are set to zeros.
    
    Parameters
    ----------
    array : array_like
        Input 2d array or image. 
    a : float or int
        Semi-major axis.
    b : float or int
        Semi-minor axis.
    PA : deg, float
        The PA of the semi-major axis.
    output_values : bool, optional
        If True returns the values of the pixels in the annulus.
    cy, cx : int or None, optional
        Coordinates of the circle center. If ``None``, the center is determined
        by the ``frame_center`` function.
    output_indices : bool, optional
        If True returns the indices inside the annulus.
    
    Returns
    -------
    Depending on output_values, output_indices:
    values : array_like
        1d array with the values of the pixels in the circular region.
    array_masked : array_like
        Input array with the circular mask applied.
    y, x : array_like
        Coordinates of pixels in circle.
    """

    if array.ndim != 2:
        raise TypeError('Input array is not a frame or 2d array.')
    sy, sx = array.shape
    if cy is None or cx is None:
        cy, cx = frame_center(array, verbose=False)

    # Definition of other parameters of the ellipse
    f = np.sqrt(a ** 2 - b ** 2)  # distance between center and foci of the ellipse
    PA_rad = np.deg2rad(PA)
    pos_f1 = (cy + f * np.cos(PA_rad), cx + f * np.sin(PA_rad))  # coords of first focus
    pos_f2 = (cy - f * np.cos(PA_rad), cx - f * np.sin(PA_rad))  # coords of second focus

    yy, xx = np.ogrid[:sy, :sx]
    # ogrid is a multidim mesh creator (faster than mgrid)
    ellipse = dist(yy, xx, pos_f1[0], pos_f1[1]) + dist(yy, xx, pos_f2[0],
                                                        pos_f2[1])
    ellipse_mask = ellipse < 2 * a  # mask of 1's and 0's

    if output_values and not output_indices:
        values = array[ellipse_mask]
        return values
    elif output_indices and not output_values:
        indices = np.array(np.where(ellipse_mask))
        y = indices[0]
        x = indices[1]
        return y, x
    elif output_indices and output_values:
        msg = 'output_values and output_indices cannot be both True.'
        raise ValueError(msg)
    else:
        array_masked = array * ellipse_mask
        return array_masked


def get_annulus_segments(array, inner_radius, width, nsegm=8, theta_init=0,
                         optim_scale_fact=1, output_values=False):
    """ Returns indices or values in segments of a centerered annulus from a
    2d ndarray.

    Parameters
    ----------
    array : array_like
        Input 2d array or image.
    inner_radius : float
        The inner radius of the donut region.
    width : float
        The size of the annulus.
    nsegm : int
        Number of segments of annulus to be extracted.
    theta_init : int
        Initial azimuth [degrees] of the first segment, counting from the postivie
        y-axis clockwise.
    optim_scale_fact : float
        Enlargen the width of the segments, which can then be used as
        optimization segments (like LOCI).
    output_values : bool, optional
        If True returns the values of the pixels in the each quadrant instead
        of the indices.

    Returns
    -------
    indices : list with lenght nsegm
        Coordinates of pixels for each annulus segment.
    If output_values is True the pixel values are returned instead.

    """
    if array.ndim != 2:
        raise TypeError('Input array is not a frame or 2d array')
    if not isinstance(nsegm, int):
        raise TypeError('`nsegm` must be an integer')

    cy, cx = frame_center(array)
    azimuth_coverage = np.deg2rad(int(np.ceil(360 / nsegm)))
    twopi = 2 * np.pi

    xx, yy = np.mgrid[:array.shape[0], :array.shape[1]]
    rad = np.sqrt((xx - cx) ** 2 + (yy - cy) ** 2)
    phi = np.arctan2(yy - cy, xx - cx)
    phirot = phi % twopi
    outer_radius = inner_radius + (width*optim_scale_fact)
    indices = []

    for i in range(nsegm):
        phi_start = np.deg2rad(theta_init) + (i * azimuth_coverage)
        phi_end = phi_start + azimuth_coverage

        if phi_start < twopi and phi_end > twopi:
            indices.append(
                np.where((rad >= inner_radius) & (rad < outer_radius) &
                         (phirot >= phi_start) & (phirot <= twopi) |
                         (rad >= inner_radius) & (rad < outer_radius) &
                         (phirot >= 0) & (phirot < phi_end - twopi)))
        elif phi_start >= twopi and phi_end > twopi:
            indices.append(
                np.where((rad >= inner_radius) & (rad < outer_radius) &
                         (phirot >= phi_start - twopi) &
                         (phirot < phi_end - twopi)))
        else:
            indices.append(
                    np.where((rad >= inner_radius) & (rad < outer_radius) &
                             (phirot >= phi_start) & (phirot < phi_end)))

    if output_values:
        values = [array[segment] for segment in indices]
        return np.array(values)
    else:
        return indices


### TODO: remove this in VIP v1.0.0. Replaced with get_annulus_segments
def get_annulus(array, inner_radius, width, output_values=False, 
                output_indices=False):                                          
    """Returns a centerered annulus from a 2d ndarray. All the rest pixels are 
    set to zeros. 
    
    Parameters
    ----------
    array : array_like
        Input 2d array or image. 
    inner_radius : float
        The inner radius of the donut region.
    width : int
        The size of the annulus.
    output_values : bool, optional
        If True returns the values of the pixels in the annulus.
    output_indices : bool, optional
        If True returns the indices inside the annulus.
    
    Returns
    -------
    Depending on output_values, output_indices:
    values : array_like
        1d array with the values of the pixels in the annulus.
    array_masked : array_like
        Input array with the annular mask applied.
    y, x : array_like
        Coordinates of pixels in annulus.
    """
    if array.ndim != 2:
        raise TypeError('Input array is not a frame or 2d array')
    array = array.copy()
    cy, cx = frame_center(array)
    yy, xx = np.mgrid[:array.shape[0], :array.shape[1]]
    circle = np.sqrt((xx - cx)**2 + (yy - cy)**2)                                                                               
    donut_mask = (circle <= (inner_radius + width)) & (circle >= inner_radius)
    if output_values and not output_indices:
        values = array[donut_mask]
        return values
    elif output_indices and not output_values:      
        indices = np.array(np.where(donut_mask))
        y = indices[0]
        x = indices[1]
        return y, x
    elif output_indices and output_values:
        raise ValueError('output_values and output_indices cannot be both True')
    else:
        array_masked = array*donut_mask
        return array_masked


### TODO: VIP v1.0.0: make use of get_annulus_segments instead
def get_annulus_cube(array, inner_radius, width, output_values=False):     
    """ Returns a centerered annulus from a 3d ndarray. All the rest pixels are 
    set to zeros. 
    
    Parameters
    ----------
    array : array_like
        Input 2d array or image. 
    inner_radius : int
        The inner radius of the donut region.
    width : int
        The size of the annulus.
    output_values : {False, True}, optional
        If True returns the values of the pixels in the annulus.
    
    Returns
    -------
    Depending on output_values:
    values : array_like
        1d array with the values of the pixels in the circular region.
    array_masked : array_like
        Input array with the annular mask applied.

    """
    if array.ndim != 3:
        raise TypeError('Input array is not a cube or 3d array.')
    arr_annulus = np.empty_like(array)
    if output_values:
        values = []
        for i in range(array.shape[0]):
            values.append(get_annulus(array[i], inner_radius, width,
                                      output_values=True))
        return np.array(values)
    else:
        for i in range(array.shape[0]):
            arr_annulus[i] = get_annulus(array[i], inner_radius, width)
        return arr_annulus
    

def get_ell_annulus(array, a, b, PA, width, output_values=False,
                    output_indices=False, cy=None, cx=None):
    """Returns a centered elliptical annulus from a 2d ndarray. All the rest 
    pixels are set to zeros. 

    Parameters
    ----------
    array : array_like
        Input 2d array or image. 
    a : flt
        Semi-major axis.
    b : flt
        Semi-minor axis.
    PA : deg
        The PA of the semi-major axis.
    width : flt
        The size of the annulus along the semi-major axis; it is proportionnally 
        thinner along the semi-minor axis).
    output_values : {False, True}, optional
        If True returns the values of the pixels in the annulus.
    output_indices : {False, True}, optional
        If True returns the indices inside the annulus.
    cy,cx: float, optional
        Location of the center of the annulus to be defined. If not provided, 
    it assumes the annuli are centered on the frame.

    Returns
    -------
    Depending on output_values, output_indices:
    values : array_like
        1d array with the values of the pixels in the annulus.
    array_masked : array_like
        Input array with the annular mask applied.
    y, x : array_like
        Coordinates of pixels in annulus.

    """
    if array.ndim != 2:
        raise TypeError('Input array is not a frame or 2d array.')
    if cy is None or cx is None:
        cy, cx = frame_center(array)
    sy, sx = array.shape

    width_a = width
    width_b = width * b / a

    # Definition of big ellipse
    f_big = np.sqrt((a + width_a / 2) ** 2 - (b + width_b / 2) ** 2)
            # distance between center and foci of the ellipse
    PA_rad = np.deg2rad(PA)
    pos_f1_big = (cy + f_big * np.cos(PA_rad),
                  cx + f_big * np.sin(PA_rad))  # coords of first focus
    pos_f2_big = (cy - f_big * np.cos(PA_rad),
                  cx - f_big * np.sin(PA_rad))  # coords of second focus

    # Definition of small ellipse
    f_sma = np.sqrt((a - width_a / 2) ** 2 - (b - width_b / 2) ** 2)
            # distance between center and foci of the ellipse
    pos_f1_sma = (cy + f_sma * np.cos(PA_rad),
                  cx + f_sma * np.sin(PA_rad))  # coords of first focus
    pos_f2_sma = (cy - f_sma * np.cos(PA_rad),
                  cx - f_sma * np.sin(PA_rad))  # coords of second focus

    yy, xx = np.ogrid[:sy, :sx]
    big_ellipse = dist(yy, xx, pos_f1_big[0], pos_f1_big[1]) + \
                  dist(yy, xx, pos_f2_big[0], pos_f2_big[ 1])
    small_ellipse = dist(yy, xx, pos_f1_sma[0], pos_f1_sma[1]) + \
                    dist(yy, xx, pos_f2_sma[0], pos_f2_sma[1])
    ell_ann_mask = ((big_ellipse < 2 * (a + width / 2)) &
                    (small_ellipse >= 2 * (a - width / 2)))# mask of 1's and 0's

    if output_values and not output_indices:
        values = array[ell_ann_mask]
        return values
    elif output_indices and not output_values:
        indices = np.array(np.where(ell_ann_mask))
        y = indices[0]
        x = indices[1]
        return y, x
    elif output_indices and output_values:
        msg = 'output_values and output_indices cannot be both True.'
        raise ValueError(msg)
    else:
        array_masked = array * ell_ann_mask
        return array_masked


def matrix_scaling(matrix, scaling):
    """ Scales a matrix using sklearn.preprocessing.scale function.

    Parameters
    ----------
    matrix : array_like
        Input 2d array.
    scaling : {None, 'temp-mean', 'spat-mean', 'temp-standard', 'spat-standard'}, str optional
        With None, no scaling is performed on the input data before SVD. With
        "temp-mean" then temporal px-wise mean subtraction is done, with
        "spat-mean" then the spatial mean is subtracted, with "temp-standard"
        temporal mean centering plus scaling to unit variance is done and with
        "spat-standard" spatial mean centering plus scaling to unit variance is
        performed.

    Returns
    -------
    matrix : array_like
        2d array with scaled values.
    """
    if scaling is None:
        pass
    elif scaling == 'temp-mean':
        matrix = scale(matrix, with_mean=True, with_std=False)
    elif scaling == 'spat-mean':
        matrix = scale(matrix, with_mean=True, with_std=False, axis=1)
    elif scaling == 'temp-standard':
        matrix = scale(matrix, with_mean=True, with_std=True)
    elif scaling == 'spat-standard':
        matrix = scale(matrix, with_mean=True, with_std=True, axis=1)
    else:
        raise ValueError('Scaling mode not recognized')

    return matrix


def prepare_matrix(array, scaling=None, mask_center_px=None, mode='fullfr',
                   annulus_radius=None, annulus_width=None, verbose=True):
    """ Builds the matrix for the SVD/PCA and other matrix decompositions,
    centers the data and masks the frames central area if needed.

    Parameters
    ----------
    array : array_like
        Input cube, 3d array.
    scaling : {None, 'temp-mean', 'spat-mean', 'temp-standard', 'spat-standard'}
        With None, no scaling is performed on the input data before SVD. With
        "temp-mean" then temporal px-wise mean subtraction is done, with
        "spat-mean" then the spatial mean is subtracted, with "temp-standard"
        temporal mean centering plus scaling to unit variance is done and with
        "spat-standard" spatial mean centering plus scaling to unit variance is
        performed.
    mask_center_px : None or Int, optional
        Whether to mask the center of the frames or not.
    mode : {'fullfr', 'annular'}, str optional
        Whether to use the whole frames or a single annulus.
    annulus_radius : float, optional
        Distance in pixels from the center of the frame to the center of the
        annulus.
    annulus_width : float, optional
        Width of the annulus in pixels.
    verbose : {True, False}, bool optional
        If True prints intermediate info and timing.

    Returns
    -------
    If mode is `annular` then the indices of the annulus (yy, xx) are returned
    along with the matrix.

    matrix : array_like
        Out matrix whose rows are vectorized frames from the input cube.

    """
    if mode == 'annular':
        if annulus_radius is None or annulus_width is None:
            raise ValueError('Annulus_radius and/or annulus_width can be None '
                             'in annular mode')

        ind = get_annulus(array[0], annulus_radius - annulus_width / 2,
                          annulus_width, output_indices=True)
        yy, xx = ind
        matrix = array[:, yy, xx]

        matrix = matrix_scaling(matrix, scaling)

        if verbose:
            msg = 'Done vectorizing the cube annulus. Matrix shape: ({}, {})'
            print(msg.format(matrix.shape[0], matrix.shape[1]))
        return matrix, ind

    elif mode == 'fullfr':
        if mask_center_px:
            array = mask_circle(array, mask_center_px)

        nfr = array.shape[0]
        matrix = np.reshape(array, (nfr, -1))  # == for i: array[i].flatten()

        matrix = matrix_scaling(matrix, scaling)

        if verbose:
            msg = 'Done vectorizing the frames. Matrix shape: ({}, {})'
            print(msg.format(matrix.shape[0], matrix.shape[1]))
        return matrix


def reshape_matrix(array, y, x):
    """ Converts a matrix whose rows are vectorized frames to a cube with
    reshaped frames.
    """
    return array.reshape(array.shape[0], y, x)

        

